import uuid
from typing import Any, Optional, Tuple

import click

from gcs_cli.config import DEFAULT_AUTH_POLICY_TIMEOUT
from gcs_cli.helpers import lookup_project, requires_login
from gcs_cli.parsing import auth_policy_create_and_update_params, globus_command
from gcs_cli.prettyprint import FORMAT_TEXT_RECORD, formatted_print
from gcs_cli.sdk_client.auth_client import MissingConsent
from gcs_cli.services.auth import get_client


@globus_command(
    "create",
    short_help="Create an authentication policy",
    disable_options=["explicit_host", "format"],
)
@click.argument("DESCRIPTION", type=click.STRING)
@click.argument("DISPLAY_NAME", type=click.STRING)
@click.option(
    "--high-assurance",
    is_flag=True,
    default=False,
    help=(
        "Flag indicating if authentication policy will be used on high assurance "
        "collections. This setting is immutable."
    ),
)
@auth_policy_create_and_update_params(create=True)
@requires_login
def auth_policy_create(
    project_id: Optional[uuid.UUID],
    high_assurance: bool,
    include: Tuple[str],
    exclude: Tuple[str],
    description: str,
    display_name: str,
    authentication_assurance_timeout: Optional[int],
) -> None:
    client = get_client()

    # If the caller did not specify a project ID, we are going to lookup projects
    # owned by the caller and, if there is only a single project, we will use it.
    if project_id is None:
        project_id = lookup_project(client, "--project-id")

    optional_fields: dict[str, Any] = {}

    for constraint, domains in [
        ("domain_constraints_include", include),
        ("domain_constraints_exclude", exclude),
    ]:
        # Remove any duplicates
        domain_list = list(set(domains))
        # For create policy, --include="" has the same effect as not using the
        # --include option at all.
        domain_list = list(filter(lambda x: x != "", domain_list))
        # For create policy, if the domain list is empty (or was ""), leave it as
        # the default of null (include all domains).
        if domain_list != []:
            optional_fields[constraint] = domain_list

    if authentication_assurance_timeout is None:
        authentication_assurance_timeout = DEFAULT_AUTH_POLICY_TIMEOUT

    try:
        policy_res = client.create_policy(
            project_id=project_id,
            display_name=display_name,
            description=description,
            high_assurance=high_assurance,
            authentication_assurance_timeout=authentication_assurance_timeout,
            **optional_fields,
        )
    except MissingConsent as e:
        raise click.ClickException(
            "Missing consent for the policy API. Please run the following "
            "command:\n"
            f"    globus-connect-server session consent {e.scope}"
        )

    formatted_print(
        policy_res["policy"],
        fields=[("Authentication Policy ID", "id")],
        text_format=FORMAT_TEXT_RECORD,
    )
