from __future__ import annotations

import typing as t

import click

from gcs_cli.helpers import requires_login
from gcs_cli.parsing import (
    auth_policy_create_and_update_params,
    auth_policy_id_arg,
    globus_command,
)
from gcs_cli.sdk_client.auth_client import MissingConsent
from gcs_cli.services.auth import get_client

if t.TYPE_CHECKING:
    import uuid


@globus_command(
    "update",
    short_help="Update an existing authentication policy",
    disable_options=["explicit_host", "format"],
)
@auth_policy_id_arg
@click.option(
    "--description",
    metavar="DESCRIPTION",
    type=click.STRING,
    help="Description for the authentication policy.",
)
@click.option(
    "--display-name",
    metavar="DISPLAY_NAME",
    type=click.STRING,
    help="Display name for the authentication policy.",
)
@auth_policy_create_and_update_params(create=False)
@requires_login
def auth_policy_update(
    auth_policy_id: uuid.UUID,
    project_id: uuid.UUID | None,
    include: t.Iterable[str],
    exclude: t.Iterable[str],
    description: str | None,
    display_name: str | None,
    authentication_assurance_timeout: int | None,
    **kwargs: t.Mapping[str, t.Any],
) -> None:
    client = get_client()

    optional_fields: dict[str, t.Any] = {}

    if project_id is not None:
        optional_fields["project_id"] = project_id
    if description is not None:
        optional_fields["description"] = description
    if display_name is not None:
        optional_fields["display_name"] = display_name
    if authentication_assurance_timeout is not None:
        optional_fields[
            "authentication_assurance_timeout"
        ] = authentication_assurance_timeout

    for constraint, domains in [
        ("domain_constraints_include", include),
        ("domain_constraints_exclude", exclude),
    ]:
        # Remove any duplicates
        domain_list = list(set(domains))
        # If the caller specified any domain(s)
        if domain_list != []:
            # Caller wants to remove all previous constraints
            if domain_list == [""]:
                optional_fields[constraint] = None
            else:
                # Remove any superfluous ''
                domain_list = list(filter(lambda x: x != "", domain_list))
                optional_fields[constraint] = domain_list

    try:
        client.update_policy(auth_policy_id, **optional_fields)
    except MissingConsent as e:
        raise click.ClickException(
            "Missing consent for the policy API. Please run the following command:\n"
            f"    globus-connect-server session consent {e.scope}"
        )
    click.echo(f"Updated authentication policy {auth_policy_id}")
