from __future__ import annotations

import uuid
from typing import Optional

import click

from globus_sdk.paging import Paginator

from gcs_cli.helpers import path_restrictions_field, requires_login, sorted_json_field
from gcs_cli.parsing import collection_id_arg, globus_command, page_size_opt
from gcs_cli.prettyprint import FORMAT_TEXT_TABLE, FieldSpec, formatted_print
from gcs_cli.services.gcs import get_client

FIELDS: FieldSpec = [
    ("Sharing Policy ID", "id"),
    ("Collection ID", "collection_id"),
    ("Sharing Path Restrictions", path_restrictions_field("sharing_restrict_paths")),
    ("Users", sorted_json_field("users")),
]


@globus_command("list", short_help="List Sharing Policies.")
@collection_id_arg
@click.option(
    "--username",
    metavar="USERNAME",
    help="Only list sharing policies which apply to the given user",
)
@page_size_opt
@requires_login
def sharing_policy_list(
    collection_id: uuid.UUID,
    username: Optional[str],
    page_size: int | None,
) -> None:
    """
    List Sharing Policies for a collection, optionally restricting those that
    apply to a particular username.
    """
    client = get_client()
    params = {}
    if page_size:
        params["page_size"] = page_size
    res: Paginator = client.paginated.get_sharing_policy_list(
        collection_id, username=username, **params
    )

    formatted_print(
        res,
        text_format=FORMAT_TEXT_TABLE,
        fields=FIELDS,
    )
