import uuid
from typing import Any, Optional, Tuple

from gcs_cli.commands.storage_gateway._common import make_update_command
from gcs_cli.parsing import (
    dropbox_options,
    globus_command,
    storage_gateway_create_and_update_params,
    storage_gateway_id_arg,
)
from gcs_cli.sdk_client.data import DropboxStoragePolicies, StorageGatewayDocument
from gcs_cli.services.gcs import DROPBOX_CONNECTOR_ID


@globus_command("dropbox", short_help="Update a Dropbox Storage Gateway")
@storage_gateway_id_arg
@storage_gateway_create_and_update_params(create=False)
@dropbox_options(create=False)
@make_update_command
def dropbox_update(
    storage_gateway_id: uuid.UUID,
    *,
    dropbox_client_id: Optional[str],
    dropbox_client_secret: Optional[str],
    dropbox_user_api_rate_limit: Optional[int],
    dropbox_allow_any_account: Optional[bool],
    **kwargs: Any
) -> Tuple[uuid.UUID, StorageGatewayDocument]:
    """
    Update an existing Dropbox Storage Gateway.
    Any options given will replace existing values in the storage gateway's definition.
    Any options not given will retain their existing values.
    """
    policies = None
    if any(
        x is not None
        for x in (
            dropbox_client_id,
            dropbox_client_secret,
            dropbox_user_api_rate_limit,
            dropbox_allow_any_account,
        )
    ):
        policies = DropboxStoragePolicies(
            client_id=dropbox_client_id,
            secret=dropbox_client_secret,
            user_api_rate_limit=dropbox_user_api_rate_limit,
            allow_any_account=dropbox_allow_any_account,
        )

    return (
        storage_gateway_id,
        StorageGatewayDocument(
            connector_id=DROPBOX_CONNECTOR_ID, policies=policies, **kwargs
        ),
    )
