import collections.abc
import uuid
from typing import Callable, Optional, Union

import click

from globus_sdk import GCSAPIError, IdentityMap

from gcs_cli.sdk_client.data import UserCredentialDocument
from gcs_cli.services.auth import get_client as get_auth_client
from gcs_cli.services.gcs import get_client


def _s3_create(
    globus_identity: Union[str, uuid.UUID],
    mapped_user: Optional[str],
    replace_existing: bool,
    s3_access_key_id: str,
    s3_secret_key: str,
    s3_requester_pays: Optional[bool],
    storage_gateway_id: str,
    connector_id: str,
    policy_class: Callable,
) -> None:
    """
    Create an S3 user credential for a particular user mapping on a storage
    gateway. By default, users can only create credentials for their own
    account mappings.  However, if the admin_managed_credentials property is
    set on a storage gateway, then endpoint administrators may create
    credentials on behalf of other users.
    """
    auth_client = get_auth_client()
    client = get_client()

    idmap = IdentityMap(auth_client)

    if isinstance(globus_identity, uuid.UUID):
        globus_identity = str(globus_identity)

    idmap.add(globus_identity)

    try:
        identity_document = idmap[globus_identity]
    except KeyError:
        raise click.ClickException(f"Unable to resolve identity {globus_identity}")

    globus_username = identity_document["username"]
    identity_id = identity_document["id"]

    if not mapped_user:
        mapped_user = globus_username

    assert mapped_user

    user_credential = UserCredentialDocument(
        connector_id=connector_id,
        identity_id=identity_id,
        storage_gateway_id=storage_gateway_id,
        username=mapped_user,
        policies=policy_class(
            s3_key_id=s3_access_key_id,
            s3_secret_key=s3_secret_key,
            s3_requester_pays=s3_requester_pays,
        ),
    )

    try:
        res = client.create_user_credential(user_credential)
        click.echo(f"Created credential {res.data['id']} for {globus_username}")
    except GCSAPIError as e:
        if e.http_status == 409:
            if (
                replace_existing
                and e.detail_data_type
                and e.detail_data_type.startswith("resource_conflict#1.")
                and isinstance(e.detail, collections.abc.Mapping)
                and e.detail.get("resources")
                and isinstance(e.detail["resources"], list)
                and len(e.detail["resources"]) == 1
            ):
                user_credential_id = e.detail["resources"][0]
                res = client.put_user_credential(user_credential_id, user_credential)
                click.echo(f"Updated credential {res.data['id']} for {globus_username}")
            else:
                raise click.ClickException(
                    f"Credential for {globus_username} already exists"
                )
        else:
            raise click.ClickException(
                f"Error creating credential for {globus_username}: {e}"
            )
