import uuid
from typing import Optional, Union

import click

from globus_sdk import GCSAPIError, IdentityMap

from gcs_cli.helpers import requires_login
from gcs_cli.parsing import UUIDorUsername, globus_command, storage_gateway_id_arg
from gcs_cli.services.auth import get_client as get_auth_client
from gcs_cli.services.gcs import get_client


@globus_command(
    "oauth-create",
    disable_options=["format"],
    short_help="Create a User Credential for connectors that support OAuth2.",
)
@click.option(
    "--globus-identity",
    type=UUIDorUsername(),
    required=True,
    help="Globus identity id or username id to associate the credential with",
)
@click.option(
    "--mapped-user",
    help="GCSv5 mapped identity username. If not provided, defaults to the Globus "
    "identity username",
)
@storage_gateway_id_arg
@requires_login
def oauth_create(
    globus_identity: Union[str, uuid.UUID],
    mapped_user: Optional[str],
    storage_gateway_id: str,
) -> None:
    """
    Create an OAuth2 user credential for a particular user mapping on a storage
    gateway. By default, users can only create credentials for their own
    account mappings.  However, if the admin_managed_credentials property is
    set on a storage gateway, then endpoint administrators may create
    credentials on behalf of other users.

    Connectors that use OAuth2 user credentials include the Azure Blob, Box,
    Dropbox, Google Drive, Google Cloud Storage, and OneDrive connectors.
    However, this command is not supported with the Azure Blob, Box, or
    Google Cloud Storage connectors when configured with enterprise or
    service mode credentials.
    """

    auth_client = get_auth_client()
    client = get_client()

    idmap = IdentityMap(auth_client)

    if isinstance(globus_identity, uuid.UUID):
        globus_identity = str(globus_identity)

    idmap.add(globus_identity)

    try:
        identity_document = idmap[globus_identity]
    except KeyError:
        raise click.ClickException(f"Unable to resolve identity {globus_identity}")

    globus_username = identity_document["username"]
    identity_id = identity_document["id"]

    if not mapped_user:
        mapped_user = globus_username

    assert mapped_user
    params = {
        "identity_id": identity_id,
        "redirect_uri": f"{client.base_url}v1/authclicomplete",
        "storage_gateway": storage_gateway_id,
        "login_hint": mapped_user,
    }
    try:
        res = client.request(
            "post",
            "/v1/user_credentials",
            data=params,
            encoding="form",
            allow_redirects=False,
        )
        if res.http_status >= 300 and res.http_status < 400:
            click.echo(
                f"To complete credential creation, visit\n{res.headers['Location']}"
            )
    except GCSAPIError as e:
        if e.http_status == 409:
            raise click.ClickException(
                f"Credential for {globus_username} already exists"
            )
        else:
            raise click.ClickException(
                f"Error creating credential for {globus_username}: {e}"
            )
