"""
Logic for using client identities.
"""
import os
from typing import Optional, cast

import click

from globus_sdk import (
    ClientCredentialsAuthorizer,
    ConfidentialAppAuthClient,
    TransferAPIError,
    TransferClient,
)

from .auth_flows import BASE_SCOPES, GCS_SCOPE_FORMAT_STRING

# Add new client env variables as needed
_CLIENT_ENV_VARS = ["GCS_CLI_CLIENT_ID", "GCS_CLI_CLIENT_SECRET", "GCS_CLI_ENDPOINT_ID"]

# Construct a string of client env variables for messages
CLIENT_ENV_VARS_STRING = " and ".join(
    [", ".join(_CLIENT_ENV_VARS[:-1]), _CLIENT_ENV_VARS[-1]]
)

_INCOMPLETE_CLIENT_ENV_VAR_MSG = (
    CLIENT_ENV_VARS_STRING + " "
    "must be set to use a client identity. Either set all environment "
    "variables or unset them to use a normal login."
)


def _get_client_creds_from_env() -> dict[str, Optional[str]]:
    """
    Returns a dict with the client env variable values.

    The keys will be lowercase versions of _CLIENT_ENV_VARS without the GCS_CLI_ prefix.
    If the env variable is not set, the key's value will be set to None.
    """
    client_env = {}
    for v in _CLIENT_ENV_VARS:
        client_env[v[len("GCS_CLI_") :].lower()] = os.getenv(v)
    return client_env


def _get_client_creds_from_env_safe() -> dict[str, str]:
    """
    Calls _get_client_creds_from_env() and casts it to non-optional values to satisfy
    mypy.

    Note: There is an implied dependency that is_client_login() has already
    been called to verify that the client environment is correct.
    """
    return cast(dict[str, str], _get_client_creds_from_env())


def is_client_login(true_if_partial: bool = False) -> bool:
    """
    Returns:
    True: All client login env variables have been set.
    False: No client login env variables have been set.

    :param true_if_partial: Return True if any client env variables have been set.
    Useful for contexts that don't want to raise the ClickException when the error
    message may be misleading for users, ie, the user shouldn't be told to fix the
    client env settings if the command doesn't accept client logins.

    Raises ClickException: Only some client login env variables have been set. Only
    raised if true_if_partial is False.

    Note: This mimics the way the Globus CLI handles this, but with an
          additional env variable (GCS_CLI_ENDPOINT_ID).
    """
    client_env = _get_client_creds_from_env_safe()

    # Make sure we don't have a fault assumption
    assert len(client_env) == len(_CLIENT_ENV_VARS)

    # Calculate the number of non-None (or empty) env values
    valid_env_cnt = len([v for v in client_env.values() if bool(v)])

    # All options are set
    if valid_env_cnt == len(_CLIENT_ENV_VARS):
        return True

    # No options are set
    if valid_env_cnt == 0:
        return False

    # Some options are set
    if true_if_partial:
        return True

    raise click.ClickException(_INCOMPLETE_CLIENT_ENV_VAR_MSG)


def get_auth_authorizer() -> ClientCredentialsAuthorizer:
    """
    Return a authorizer based on the client credentials in the environment.

    Note: There is an implied dependency that is_client_login() has already been
    called to verify that the client environment is correct.
    """
    client_env = _get_client_creds_from_env_safe()
    return ClientCredentialsAuthorizer(
        ConfidentialAppAuthClient(
            client_env["client_id"],
            client_env["client_secret"],
        ),
        BASE_SCOPES,
    )


def get_gcs_authorizer(gcs_client_id: str) -> ClientCredentialsAuthorizer:
    """
    Return an authorizer based on the client credentials in the environment.

    Note: There is an implied dependency that is_client_login() has already been
    called to verify that the client environment is correct.
    """
    client_env = _get_client_creds_from_env_safe()
    return ClientCredentialsAuthorizer(
        ConfidentialAppAuthClient(
            client_env["client_id"],
            client_env["client_secret"],
        ),
        GCS_SCOPE_FORMAT_STRING.format(gcs_client_id),
    )


def lookup_endpoint_info() -> tuple[str, str]:
    """
    Get the endpoint's FQDN and GCS client ID from Transfer. The GCS client ID may
    not match the endpoint ID if this is a GCSv5.3 endpoint migrated to GCSv5.4.

    Since the endpoint_id is stuffed in the environment as part of the client
    login, the caller does not need to provide it.

    Note: There is an implied dependency that is_client_login() has already been
    called to verify that the client environment is correct.
    """
    client_env = _get_client_creds_from_env_safe()
    authorizer = ClientCredentialsAuthorizer(
        ConfidentialAppAuthClient(
            client_env["client_id"],
            client_env["client_secret"],
        ),
        "urn:globus:auth:scope:transfer.api.globus.org:all",
    )
    transfer_client = TransferClient(authorizer=authorizer)

    try:
        endpoint_info = transfer_client.get_endpoint(client_env["endpoint_id"])
    except TransferAPIError:
        raise click.ClickException(
            f"Either endpoint {client_env['endpoint_id']} does not exist or it is "
            "private\n"
        )

    gcs_client_id = endpoint_info["owner_id"]
    gcs_fqdn = endpoint_info["DATA"][0]["hostname"]
    return gcs_client_id, gcs_fqdn
