import re
from typing import Callable, Optional, Union

from cryptography import x509
from cryptography.hazmat.primitives import serialization


def load_certificate(path: str) -> str:
    """
    Load the X.509 certificate from path, returning a normalized output
    containing just the PEM-encoded data

    :param path:
        Path to the certificate file.
    """
    with open(path, "r") as f:
        data = f.read()
    cert = x509.load_pem_x509_certificate(data.encode())
    return cert.public_bytes(serialization.Encoding.PEM).decode()


def load_certificate_chain(path: str) -> Optional[str]:
    """
    Load a list of X.509 certificates from path, returning a normalized output
    containing just the PEM-encoded data

    :param path:
        Path to the certificate chain file.
    """
    if path:
        certs = []
        start = r"-----BEGIN CERTIFICATE-----[\r]?\n"
        end = r"-----END CERTIFICATE-----[\r]?[\n]?"
        with open(path, "r") as f:
            data = f.read()
        while data:
            startp = re.search(start, data, re.MULTILINE)
            endp = re.search(end, data, re.MULTILINE)
            if startp and endp and startp.start() < endp.end():
                cert = x509.load_pem_x509_certificate(
                    data[startp.start() : endp.end()].encode(),
                )
                certs.append(cert.public_bytes(serialization.Encoding.PEM).decode())
                data = data[endp.end() :]
            else:
                break
        return "\n".join(certs)
    return None


def load_private_key(
    path: str, passphrase: Optional[Union[str, Callable[..., bytes]]] = None
) -> str:
    """
    Load a private key from path, returning a normalized output
    containing just the PEM-encoded data, using the optional passphrase to
    decrypt the data if needed.

    :param path:
        Path to the private key file.
    :param passphrase:
        String or callable to return a passphrase
    """
    with open(path, "r") as f:
        data = f.read()

    if isinstance(passphrase, str):
        pp_arg = passphrase.encode()
    else:
        pp_arg = None

    try:
        private_key = serialization.load_pem_private_key(data.encode(), pp_arg)
    except ValueError:
        if callable(passphrase):
            private_key = serialization.load_pem_private_key(
                data.encode(), passphrase()
            )
    return private_key.private_bytes(
        serialization.Encoding.PEM,
        serialization.PrivateFormat.PKCS8,
        serialization.NoEncryption(),
    ).decode()
