from __future__ import annotations

import typing as t

import click


def _max_is_less_than_preferred(doc: dict[str, t.Any], opt: str) -> bool:
    max_value = doc.get(f"max_{opt}")
    preferred_value = doc.get(f"preferred_{opt}")
    if max_value is not None and preferred_value is not None:
        if int(max_value) < int(preferred_value):
            return True
    return False


def validate_network_use(doc: dict[str, t.Any], create: bool = False) -> None:
    """
    Validate network use parameters. A few validations are done:

    If create is true, and network_use is custom, all of max_parallelism,
    preferred_parallelism, max_concurrency, and preferred_concurrency must be
    set.

    If create is true and network_use is not custom, none of the other network
    params are allowed.

    If network_use is present and not custom, none of max_parallelism,
    preferred_parallelism, max_concurrency, or preferred_concurrency may be
    set

    If max_parallelism and preferred_parallelism are both present, max must be
    >= preferred.

    If max_concurrency and preferred_concurrency are both present, max must
    be >= preferred

    Any failure with throw an exception.
    """

    network_params = [
        "max_parallelism",
        "preferred_parallelism",
        "max_concurrency",
        "preferred_concurrency",
    ]
    network_use = doc.get("network_use")
    if create and network_use is not None:
        if network_use == "custom":
            for n in network_params:
                if doc.get(n) is None:
                    n_option = f"--{n.replace('_', '-')}"
                    raise click.UsageError(
                        f"{n_option} required when setting network_use to custom"
                    )
    elif create:
        for n in network_params:
            if doc.get(n) is None:
                n_option = f"--{n.replace('_', '-')}"
                raise click.UsageError(
                    f"{n_option} not allowed when not setting network_use to custom"
                )
    if network_use is not None and network_use != "custom":
        for n in network_params:
            if doc.get(n) is not None:
                n_option = f"--{n.replace('_', '-')}"
                raise click.UsageError(
                    f"{n_option} not allowed when not setting network_use to custom"
                )

    for opt in ["parallelism", "concurrency"]:
        if _max_is_less_than_preferred(doc, opt):
            raise click.UsageError(
                f"--max-{opt} must not be less than --preferred-{opt}"
            )
