"""
Data helper classes for constructing GCS API documents. All classes should
extend ``dict``, so they can be passed seamlessly to
:class:`GCSClient` methods without conversion.
"""
from __future__ import annotations

import typing as t

from gcs_cli.parsing.types import _ExplicitNullClass

if t.TYPE_CHECKING:
    import uuid
    from collections import abc

    from typing_extensions import Literal


def _set_raw_kwargs(d: abc.MutableMapping, **kwargs: t.Any) -> None:
    for arg, val in kwargs.items():
        if val is not None:
            d[arg] = val


def _set_bool_kwargs(d: abc.MutableMapping, **kwargs: t.Any) -> None:
    for arg, val in kwargs.items():
        if val is not None:
            d[arg] = bool(val)


def _set_string_kwargs(d: abc.MutableMapping, **kwargs: t.Any) -> None:
    for arg, val in kwargs.items():
        if val is not None:
            d[arg] = str(val)


def _set_string_list_kwargs(d: abc.MutableMapping, **kwargs: t.Any) -> None:
    for arg, val in kwargs.items():
        if val is not None:
            d[arg] = [str(x) for x in val]


class StorageGatewayDocument(dict):
    """
    Convenience class for constructing a Storage Gateway document
    to use as the `data` parameter to :meth:`create_storage_gateway` or
    :meth:`update_storage_gateway`

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``display_name`` (*string*)
      Name of the Storage Gateway.

    ``connector_id`` (*string*)
      UUID of the connector type that this Storage Gateway interacts with.

    ``identity_mappings`` (*list of dicts*)
      A list of IdentityMapping objects which are applied to user identities to attempt
      to determine what accounts are available for access.

    ``policies`` (*dict*)
      Connector specific storage policies. It is recommended that
      you use one of the Policy helper classes (e.g. `POSIXStoragePolicies`
      if you are using the POSIX connector) to create these.

    ``allowed_domains`` (*list of string*)
      List of allowed domains. Users creating credentials or collections on
      this Storage Gateway must have an identity in one of these domains.

    ``high_assurance`` (*bool*)
      Flag indicating if the Storage Gateway requires high assurance features.

    ``require_mfa`` (*bool*)
      Flag indicating if the Storage Gateway requires multi-factor authentication.

    ``authentication_timeout_mins`` (*int*)
      Timeout (in minutes) during which a user is required to have authenticated
      in a session to access this storage gateway.

    ``restrict_paths`` (*dict*)
      Path restrictions within this Storage Gateway. [Private]

    ``users_allow`` (*list of string*)
      List of local users explicitly allowed access to this storage gateway.

    ``users_deny`` (*list of string*)
      List of local users explicitly denied access to this storage gateway.

    ``admin_managed_credentials`` (*bool*)
      Flag indicating whether an endpoint administrator can create or update
      credentials owned by other users.
    """

    def __init__(
        self,
        DATA_TYPE: str = "storage_gateway#1.2.0",
        display_name: str | None = None,
        connector_id: str | None = None,
        identity_mappings: abc.Iterable[abc.Mapping] | None = None,
        policies: abc.Mapping | None = None,
        allowed_domains: abc.Iterable[str] | None = None,
        high_assurance: bool | None = None,
        require_mfa: bool | None = None,
        authentication_timeout_mins: int | None = None,
        users_allow: abc.Iterable[str] | None = None,
        users_deny: abc.Iterable[str] | None = None,
        admin_managed_credentials: bool | None = None,
        **kwargs: t.Any,
    ) -> None:
        _set_string_kwargs(
            self,
            DATA_TYPE=DATA_TYPE,
            display_name=display_name,
            connector_id=connector_id,
        )
        _set_raw_kwargs(
            self,
            identity_mappings=identity_mappings,
            policies=policies,
            high_assurance=high_assurance,
            require_mfa=require_mfa,
            authentication_timeout_mins=authentication_timeout_mins,
            users_allow=users_allow,
            users_deny=users_deny,
            admin_managed_credentials=admin_managed_credentials,
        )
        _set_string_list_kwargs(self, allowed_domains=allowed_domains)

        self.update(kwargs)


class POSIXStoragePolicies(dict):
    """
    Convenience class for constructing a POSIX Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``groups_allow`` (*list of string*)
      List of POSIX group IDs allowed to access this Storage Gateway. [Private]

    ``groups_deny`` (*list of string*)
      List of POSIX group IDs denied access this Storage Gateway. [Private]
    """

    def __init__(
        self,
        DATA_TYPE: str = "posix_storage_policies#1.0.0",
        groups_allow: abc.Iterable[str] | None = None,
        groups_deny: abc.Iterable[str] | None = None,
        **kwargs: t.Any,
    ):
        self["DATA_TYPE"] = str(DATA_TYPE)

        _set_string_list_kwargs(
            self, groups_allow=groups_allow, groups_deny=groups_deny
        )

        self.update(kwargs)


class POSIXStagingStoragePolicies(POSIXStoragePolicies):
    """
    Convenience class for constructing a POSIX Staging Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``groups_allow`` (*list of string*)
      List of POSIX group IDs allowed to access this Storage Gateway. [Private]

    ``groups_deny`` (*list of string*)
      List of POSIX group IDs denied access this Storage Gateway. [Private]

    ``stage_app`` (*string*)
      Path to the stage app.[Private]

    ``environment`` (*list of string*)
      Variables to set in the environment when executing the stage_app.[Private]
    """

    def __init__(
        self,
        DATA_TYPE: str = "posix_staging_storage_policies#1.0.0",
        stage_app: str | None = None,
        environment: abc.Iterable[abc.Mapping[str, str]] | None = None,
        **kwargs: t.Any,
    ):
        if environment is not None:
            self["environment"] = [{**e} for e in environment]

        if stage_app is not None:
            _set_string_kwargs(self, stage_app=stage_app)

        super().__init__(DATA_TYPE=DATA_TYPE, **kwargs)


class BlackPearlStoragePolicies(dict):
    """
    Convenience class for constructing a BlackPearl Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``s3_endpoint`` (*string*)
      The URL of the S3 endpoint of the BlackPearl appliance to use to access
      collections on this Storage Gateway.

    ``bp_access_id_file`` (*string*)
      Path to the file which provides mappings from usernames within the
      configured identity domain to the ID and secret associated with the user's
      BlackPearl account
    """

    def __init__(
        self,
        DATA_TYPE: str = "blackpearl_storage_policies#1.0.0",
        s3_endpoint: str | None = None,
        bp_access_id_file: str | None = None,
        **kwargs: t.Any,
    ) -> None:
        _set_string_kwargs(
            self,
            DATA_TYPE=DATA_TYPE,
            s3_endpoint=s3_endpoint,
            bp_access_id_file=bp_access_id_file,
        )

        self.update(kwargs)


class BoxStoragePolicies(dict):
    """
    Convenience class for constructing a Box Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``enterpriseID`` (*string*)
      Identifies which Box Enterprise this Storage Gateway is authorized access
      to. [Private]

    ``boxAppSettings`` (*dict*)
      Values that the Storage Gateway uses to identify and authenticate the
      with the Box API. [Private]

    ``client_id`` (*string*)
      Box OAuth App client ID. [Private]

    ``secret`` (*string*)
      Box OAuth App client secret. [Private]

    ``user_api_rate_limit`` (*int*)
      Maximum number of Box API calls allowed per second per user.

    ``allow_any_account`` (*bool*)
      Do not require OAuth username/email to match identity mapping.
    """

    def __init__(
        self,
        DATA_TYPE: str | None = None,
        enterpriseID: str | None = None,
        boxAppSettings: abc.Mapping | None = None,
        client_id: str | None = None,
        secret: str | None = None,
        user_api_rate_limit: int | None = None,
        allow_any_account: bool | None = None,
        **kwargs: t.Any,
    ) -> None:
        if DATA_TYPE is None:
            if allow_any_account is not None:
                DATA_TYPE = "box_storage_policies#1.2.0"
            elif client_id or secret or user_api_rate_limit:
                DATA_TYPE = "box_storage_policies#1.1.0"
            else:
                DATA_TYPE = "box_storage_policies#1.0.0"

        _set_string_kwargs(
            self,
            DATA_TYPE=DATA_TYPE,
            enterpriseID=enterpriseID,
            client_id=client_id,
            secret=secret,
        )
        _set_bool_kwargs(self, allow_any_account=allow_any_account)
        _set_raw_kwargs(
            self, boxAppSettings=boxAppSettings, user_api_rate_limit=user_api_rate_limit
        )

        self.update(kwargs)


class CephStoragePolicies(dict):
    """
    Convenience class for constructing a Ceph Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``s3_endpoint`` (*string*)
      URL of the S3 API endpoint

    ``s3_buckets`` (*list of string*)
      List of buckets not owned by the collection owner that will be shown in
      the root of collections created at the base of this Storage Gateway.

    ``ceph_admin_key_id`` (*string*)
      Administrator key id used to authenticate with the ceph admin service to
      obtain user credentials. [Private]

    ``ceph_admin_secret_key`` (*string*)
      Administrator secret key used to authenticate with the ceph admin service
      to obtain user credentials. [Private]
    """

    def __init__(
        self,
        DATA_TYPE: str = "ceph_storage_policies#1.0.0",
        s3_endpoint: str | None = None,
        s3_buckets: abc.Iterable[str] | None = None,
        ceph_admin_key_id: str | None = None,
        ceph_admin_secret_key: str | None = None,
        **kwargs: t.Any,
    ) -> None:
        _set_string_kwargs(
            self,
            DATA_TYPE=DATA_TYPE,
            s3_endpoint=s3_endpoint,
            ceph_admin_key_id=ceph_admin_key_id,
            ceph_admin_secret_key=ceph_admin_secret_key,
        )
        _set_string_list_kwargs(self, s3_buckets=s3_buckets)

        self.update(kwargs)


class DropboxStoragePolicies(dict):
    """
    Convenience class for constructing a Dropbox Storage Policy document
    to use as the `policies` parameter when creating a UserCredential

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``client_id`` (*string*)
      Client ID registered with the Dropbox App console. [Private]

    ``secret`` (*string*)
      App secret associated with the client_id. [Private]

    ``user_api_rate_limit`` (*integer*)
      API Rate limit associated with this client ID.

    ``allow_any_account`` (*bool*)
      Do not require OAuth username/email to match identity mapping.
    """

    def __init__(
        self,
        DATA_TYPE: str = "dropbox_storage_policies#1.0.0",
        client_id: str | None = None,
        secret: str | None = None,
        user_api_rate_limit: int | None = None,
        allow_any_account: bool | None = None,
        **kwargs: t.Any,
    ) -> None:
        _set_string_kwargs(
            self, DATA_TYPE=DATA_TYPE, client_id=client_id, secret=secret
        )
        _set_bool_kwargs(self, allow_any_account=allow_any_account)
        _set_raw_kwargs(self, user_api_rate_limit=user_api_rate_limit)

        self.update(kwargs)


class GoogleDriveStoragePolicies(dict):
    """
    Convenience class for constructing a Google Drive Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``client_id`` (*string*)
      Client ID registered with the Google Application console to access Google
      Drive. [Private]

    ``secret`` (*string*)
      Secret created to access access Google Drive with the client_id in this
      policy. [Private]

    ``user_api_rate_quota`` (*integer*)
      User API Rate quota associated with this client ID.

    ``allow_any_account`` (*bool*)
      Do not require OAuth username/email to match identity mapping.
    """

    def __init__(
        self,
        DATA_TYPE: str | None = None,
        client_id: str | None = None,
        secret: str | None = None,
        user_api_rate_quota: int | None = None,
        allow_any_account: bool | None = None,
        **kwargs: t.Any,
    ) -> None:
        if DATA_TYPE is None:
            if allow_any_account is not None:
                DATA_TYPE = "google_drive_storage_policies#1.1.0"
            else:
                DATA_TYPE = "google_drive_storage_policies#1.0.0"

        _set_string_kwargs(
            self, DATA_TYPE=DATA_TYPE, client_id=client_id, secret=secret
        )
        _set_bool_kwargs(self, allow_any_account=allow_any_account)
        _set_raw_kwargs(self, user_api_rate_quota=user_api_rate_quota)

        self.update(kwargs)


class GoogleCloudStoragePolicies(dict):
    """
    Convenience class for constructing a Google Cloud Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``client_id`` (*string*)
      Client ID registered with the Google Application console to access Google
      Drive. [Private]

    ``secret`` (*string*)
      Secret created to access access Google Drive with the client_id in this
      policy. [Private]

    ``service_account_key`` (*dict*)
      Credentials for use with service account auth, read from a
      Google-provided json file. [Private]

    ``buckets`` (*list of string*)
      The list of Google Cloud Storage buckets which the Storage Gateway is
      allowed to access, as well as the list of buckets that will be shown in
      root level directory listings. If this list is unset, bucket access is
      unrestricted and all non public credential accessible buckets will be
      shown in root level directory listings. The value is a list of bucket
      names.

    ``projects`` (*list of string*)
      The list of Google Cloud Storage projects which the Storage Gateway is
      allowed to access. If this list is unset, project access is unrestricted.
      The value is a list of project id strings.

    ``allow_any_account`` (*bool*)
      Do not require OAuth username/email to match identity mapping.
    """

    def __init__(
        self,
        DATA_TYPE: str | None = None,
        client_id: str | None = None,
        secret: str | None = None,
        service_account_key: abc.Mapping | None = None,
        buckets: abc.Iterable[str] | None = None,
        projects: abc.Iterable[str] | None = None,
        allow_any_account: bool | None = None,
        **kwargs: t.Any,
    ) -> None:
        if DATA_TYPE is None:
            if allow_any_account is not None:
                DATA_TYPE = "google_cloud_storage_policies#1.1.0"
            else:
                DATA_TYPE = "google_cloud_storage_policies#1.0.0"

        _set_string_kwargs(
            self, DATA_TYPE=DATA_TYPE, client_id=client_id, secret=secret
        )
        _set_string_list_kwargs(self, buckets=buckets, projects=projects)
        _set_bool_kwargs(self, allow_any_account=allow_any_account)
        _set_raw_kwargs(self, service_account_key=service_account_key)

        self.update(kwargs)


class OneDriveStoragePolicies(dict):
    """
    Convenience class for constructing a OneDrive Storage Policy document
    to use as the `policies` parameter when creating a UserCredential

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``client_id`` (*string*)
      Client ID registered with the MS Application console to access
      OneDrive. [Private]

    ``secret`` (*string*)
      Secret created to access access MS with the client_id in this
      policy. [Private]

    ``tenant`` (*string*)
      MS Tenant ID from which to allow user logins. [Private]

    ``user_api_rate_limit`` (*integer*)
      User API Rate limit associated with this client ID.

    ``allow_any_account`` (*bool*)
      Do not require OAuth username/email to match identity mapping.
    """

    def __init__(
        self,
        DATA_TYPE: str | None = None,
        client_id: str | None = None,
        secret: str | None = None,
        tenant: t.Union[str, _ExplicitNullClass] | None = None,
        user_api_rate_limit: int | None = None,
        allow_any_account: bool | None = None,
        **kwargs: t.Any,
    ) -> None:
        if DATA_TYPE is None:
            if allow_any_account is not None:
                DATA_TYPE = "onedrive_storage_policies#1.1.0"
            else:
                DATA_TYPE = "onedrive_storage_policies#1.0.0"

        _set_string_kwargs(
            self, DATA_TYPE=DATA_TYPE, client_id=client_id, secret=secret, tenant=tenant
        )
        _set_bool_kwargs(self, allow_any_account=allow_any_account)
        _set_raw_kwargs(self, user_api_rate_limit=user_api_rate_limit)

        self.update(kwargs)


class AzureBlobStoragePolicies(dict):
    """
    Convenience class for constructing a Azure Blob Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``client_id`` (*string*)
      Client ID registered with the MS Application console to access
      Azure Blob [Private]

    ``secret`` (*string*)
      Secret created to access access MS with the client_id in this
      policy. [Private]

    ``tenant`` (*string*)
      MS Tenant ID from which to allow user logins. [Private]

    ``account`` (*string*)
      Azure Storage account. [Private]

    ``auth_type`` (*string*)
      Auth type: user, service_principal or user_service_principal

    ``adls`` (*bool*)
      ADLS support enabled or not. [Private]

    ``allow_any_account`` (*bool*)
      Do not require OAuth username/email to match identity mapping.
    """

    def __init__(
        self,
        DATA_TYPE: str | None = None,
        client_id: str | None = None,
        secret: str | None = None,
        tenant: str | None = None,
        account: str | None = None,
        auth_type: str | None = None,
        adls: bool | None = None,
        allow_any_account: bool | None = None,
        **kwargs: t.Any,
    ) -> None:
        if DATA_TYPE is None:
            if allow_any_account is not None:
                DATA_TYPE = "azure_blob_storage_policies#1.1.0"
            else:
                DATA_TYPE = "azure_blob_storage_policies#1.0.0"

        _set_string_kwargs(
            self,
            DATA_TYPE=DATA_TYPE,
            client_id=client_id,
            secret=secret,
            tenant=tenant,
            account=account,
            auth_type=auth_type,
        )
        _set_bool_kwargs(self, allow_any_account=allow_any_account)
        _set_raw_kwargs(self, adls=adls)

        self.update(kwargs)


class S3StoragePolicies(dict):
    """
    Convenience class for constructing a Google Cloud Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``s3_endpoint`` (*string*)
      URL of the S3 API endpoint

    ``s3_buckets`` (*list of string*)
      List of buckets not owned by the collection owner that will be shown in
      the root of collections created at the base of this Storage Gateway.

    ``s3_user_credential_required`` (*bool*)
      Flag indicating if a Globus User must register a user credential in order
      to create a Guest Collection on this Storage Gateway.

    ``s3_requester_pays`` (*bool*)
      Flag indicating that operations on this gateway will use Requester Pays.
      Globus Users must consent to requester pays when creating the credential.
    """

    def __init__(
        self,
        DATA_TYPE: str | None = None,
        s3_endpoint: str | None = None,
        s3_buckets: abc.Iterable[str] | None = None,
        s3_user_credential_required: bool | None = None,
        s3_requester_pays: bool | None = None,
        **kwargs: t.Any,
    ) -> None:
        if DATA_TYPE is None:
            if s3_requester_pays is not None:
                DATA_TYPE = "s3_storage_policies#1.1.0"
            else:
                DATA_TYPE = "s3_storage_policies#1.0.0"

        # sanitize s3 endpoint
        while s3_endpoint and s3_endpoint[-1] == "/":
            s3_endpoint = s3_endpoint[:-1]
        _set_string_kwargs(self, DATA_TYPE=DATA_TYPE, s3_endpoint=s3_endpoint)
        _set_bool_kwargs(self, s3_requester_pays=s3_requester_pays)
        _set_raw_kwargs(
            self,
            s3_user_credential_required=s3_user_credential_required,
        )
        _set_string_list_kwargs(self, s3_buckets=s3_buckets)

        self.update(kwargs)


class ActiveScaleStoragePolicies(S3StoragePolicies):
    """
    ActiveScale Storage Policy uses S3 policy DATA_TYPE
    """

    pass


class IrodsStoragePolicies(dict):
    """
    Convenience class for constructing an iRODS Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``irods_environment_file`` (*string*)
      Path to iRODS environent file on the endpoint

    ``irods_authentication_file`` (*string*)
      Path to iRODS authentication file on the endpoint
    """

    def __init__(
        self,
        DATA_TYPE: str = "irods_storage_policies#1.0.0",
        irods_environment_file: str | None = None,
        irods_authentication_file: str | None = None,
        **kwargs: t.Any,
    ) -> None:
        _set_string_kwargs(
            self,
            DATA_TYPE=DATA_TYPE,
            irods_environment_file=irods_environment_file,
            irods_authentication_file=irods_authentication_file,
        )

        self.update(kwargs)


class HPSSStoragePolicies(dict):
    """
    Convenience class for constructing a HPSS Storage Policy document
    to use as the `policies` parameter when creating a StorageGatewayDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``authentication_mech`` (*string*)
      Authentication mechanism to use with HPSS.

    ``authenticator`` (*string*)
      Authentication credentials to use with HPSS.

    ``uda_checksum_support`` (*bool*)
      Flag indicating whether checksums should be stored in metadata.

    ``login_name`` (*string*)
      HPSS privileged account used for authentication to HPSS.
    """

    def __init__(
        self,
        DATA_TYPE: str = "hpss_storage_policies#1.0.0",
        authentication_mech: str | None = None,
        authenticator: str | None = None,
        uda_checksum_support: str | None = None,
        login_name: str | None = None,
        **kwargs: t.Any,
    ) -> None:
        if login_name is not None and DATA_TYPE == "hpss_storage_policies#1.0.0":
            DATA_TYPE = "hpss_storage_policies#1.1.0"

        _set_string_kwargs(
            self,
            DATA_TYPE=DATA_TYPE,
            authentication_mech=authentication_mech,
            authenticator=authenticator,
            login_name=login_name,
        )
        _set_bool_kwargs(
            self,
            uda_checksum_support=uda_checksum_support,
        )

        self.update(kwargs)


class CollectionDocument(dict):
    """
    Convenience class for constructing a Collection document
    to use as the `data` parameter to :meth:`update_collection`

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``storage_gateway_id`` (*string*)
      Unique ID of the Storage Gateway which this collection provides access to.
      This value can not change after the collection is created.

    ``identity_id`` (*string*)
      Globus Auth identity to who acts as the owner of this collection. This
      identity is an `administrator` on the collection.

    ``collection_type`` (*string*)
      Either "guest" or "mapped". A mapped collection requires an account on the
      system to access the administrator-defined collection. A guest collection
      allows users to share access to their data on a Storage Gateway by
      registering an credential with the GCS Manager.

    ``display_name`` (*string*)
      Friendly name for the collection. Unicode string, max 128 characters,
      no new lines (`\r` or `\n`).

    ``root_path`` (*string*)
      Root path for the collection, relative to the Storage Gateway. [Private]

    ``default_directory`` (*string*)
      Default directory when using the collection, relative to the root_path.

    ``public`` (*bool*)
      Flag indicating whether this collection is visible to other Globus users.

    ``force_encryption`` (*bool*)
      Flag indicating whether all data transfers to and from this collection
      are always encrypted.

    ``polices`` (*dict*)
      Connector specific collection policies. Most connectors do not use
      these. It is recommended that you use a CollectionPolicy helper classes
      (e.g. `GoogleCloudStorageCollectionPolicies` if you are using the
      Google Cloud Storage connector) to create these.

    ``organization`` (*string*)
      Organization that maintains the collection.

    ``department`` (*string*)
      Department within organization that runs the server(s).

    ``keywords`` (*list of strings*)
      List of search keywords for the endpoint

    ``description`` (*string*)
      A description of the endpoint.

    ``contact_email`` (*string*)
      Email address of the support contact for the collection.

    ``contact_info`` (*string*)
      Other non-email contact information for the collection, e.g. phone and
      mailing address.

    ``info_link``  (*string*)
      Link to a web page with more information about the endpoint.

    ``allow_guest_collections`` (*bool*)
      Flag indicating if this collection allows users to create guest
      collections on it. It may only be set to True on mapped collections.

    ``guest_auth_policy_id`` (*string*)
      Globus Auth policy ID to set on a mapped collection which is then inherited
      by its guest collections.

    ``domain_name`` (*string*)
      Host name label for this collection (mapped only).

    ``sharing_restrict_paths`` (*dict*)
      Restrictions on which paths may be shared in guest
      collections related to this mapped collection [private].

    ``enable_https`` (*bool*)
      Boolean flag indicating whether this collection should support HTTPS or
      not.

    ``user_message`` (*str*)
      A message for clients to display to users when interacting with this
      collection.

    ``user_message_link`` (*str*)
      Link to additional messaging for clients to display to users when
      interacting with this endpoint, linked to an http or https URL.

    ``acl_expiration_mis`` (*int*)
      Length of time that guest collection permissions are valid.
      Only settable on HA mapped collections and used by the guest
      collections attached to it.
    """

    def __init__(
        self,
        DATA_TYPE: str | None = None,
        storage_gateway_id: t.Union[uuid.UUID, str] | None = None,
        identity_id: t.Union[uuid.UUID, str] | None = None,
        collection_type: t.Union[Literal["guest"], Literal["mapped"]] | None = None,
        display_name: str | None = None,
        root_path: str | None = None,
        default_directory: str | None = None,
        policies: abc.Mapping | None = None,
        public: bool | None = None,
        force_encryption: bool | None = None,
        allow_guest_collections: bool | None = None,
        disable_anonymous_writes: bool | None = None,
        organization: str | None = None,
        department: str | None = None,
        keywords: abc.Iterable[str] | None = None,
        description: str | None = None,
        contact_email: str | None = None,
        contact_info: str | None = None,
        info_link: str | None = None,
        domain_name: str | None = None,
        **kwargs: t.Any,
    ) -> None:
        if DATA_TYPE is None:
            if "acl_expiration_mins" in kwargs:
                DATA_TYPE = "collection#1.10.0"
            elif "delete_protected" in kwargs:
                DATA_TYPE = "collection#1.8.0"
            elif (
                "user_message" in kwargs
                and kwargs["user_message"]
                and len(kwargs["user_message"]) > 64
            ):
                DATA_TYPE = "collection#1.7.0"
            elif "guest_auth_policy_id" in kwargs:
                DATA_TYPE = "collection#1.6.0"
            elif disable_anonymous_writes is not None:
                DATA_TYPE = "collection#1.5.0"
            elif "force_verify" in kwargs:
                DATA_TYPE = "collection#1.4.0"
            elif "sharing_users_allow" in kwargs or "sharing_users_deny" in kwargs:
                DATA_TYPE = "collection#1.2.0"
            elif (
                kwargs.get("enable_https") is not None
                or "user_message" in kwargs
                or "user_message_link" in kwargs
            ):
                DATA_TYPE = "collection#1.1.0"
            else:
                DATA_TYPE = "collection#1.0.0"

        _set_string_kwargs(
            self,
            DATA_TYPE=DATA_TYPE,
            storage_gateway_id=storage_gateway_id,
            identity_id=identity_id,
            collection_type=collection_type,
            display_name=display_name,
            root_path=root_path,
            default_directory=default_directory,
            organization=organization,
            department=department,
            description=description,
            contact_email=contact_email,
            contact_info=contact_info,
            info_link=info_link,
            domain_name=domain_name,
        )

        _set_bool_kwargs(
            self,
            allow_guest_collections=allow_guest_collections,
            force_encryption=force_encryption,
            public=public,
            disable_anonymous_writes=disable_anonymous_writes,
        )
        _set_raw_kwargs(self, keywords=keywords, policies=policies)

        self.update(kwargs)


class GoogleCloudStorageCollectionPolicies(dict):
    """
    Convenience class for constructing a Google Cloud Storage Collection Policy
    document to use as the `policies` parameter when creating a
    CollectionDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``project`` (*string*)
      Google Cloud Platform project ID that is used by this collection
    """

    def __init__(
        self,
        DATA_TYPE: str = "google_cloud_storage_collection_policies#1.0.0",
        project: str | None = None,
        **kwargs: t.Any,
    ) -> None:
        self["DATA_TYPE"] = str(DATA_TYPE)
        _set_string_kwargs(self, project=project)

        self.update(kwargs)


class POSIXCollectionPolicies(dict):
    """
    Convenience class for constructing a Posix Collection Policy
    document to use as the `policies` parameter when creating a
    CollectionDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``sharing_groups_allow`` (*list of strings*)
      POSIX groups which are allowed to create guest collections.

    ``sharing_groups_deny`` (*list of strings*)
      POSIX groups which are not allowed to create guest collections.
    """

    def __init__(self, **kwargs: t.Any) -> None:
        self["DATA_TYPE"] = kwargs.get("DATA_TYPE", "posix_collection_policies#1.1.0")

        if "sharing_groups_allow" in kwargs:
            _set_string_list_kwargs(
                self, sharing_groups_allow=kwargs["sharing_groups_allow"]
            )
        if "sharing_groups_deny" in kwargs:
            _set_string_list_kwargs(
                self, sharing_groups_deny=kwargs["sharing_groups_deny"]
            )


class POSIXStagingCollectionPolicies(POSIXCollectionPolicies):
    """
    Convenience class for constructing a Posix Staging Collection Policy
    document to use as the `policies` parameter when creating a
    CollectionDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``sharing_groups_allow`` (*list of strings*)
      POSIX groups which are allowed to create guest collections.

    ``sharing_groups_deny`` (*list of strings*)
      POSIX groups which are not allowed to create guest collections.
    """

    def __init__(self, **kwargs: t.Any) -> None:
        if "DATA_TYPE" not in kwargs:
            kwargs["DATA_TYPE"] = "posix_staging_collection_policies#1.0.0"
        super().__init__(**kwargs)


class RoleDocument(dict):
    """
    Convenience class for constructing a Role document
    to use as the `data` parameter to :meth:`create_role`

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``role`` (*string*)
      Role assigned to the principal. Known values are owner,
      administrator, access_manager, activity_manager,
      and activity_monitor

    ``principal`` (*string*)
      Auth identity or group id URN. Should be in the format
      urn:globus:auth:[identity|group]:{uuid of identity or group}

    ``collection`` (*string*)
      Collection ID for the collection the role will apply to.
      This value is omitted when creating an endpoint
      role or when creating role definitions when creating collections.
    """

    def __init__(
        self,
        DATA_TYPE: str = "role#1.0.0",
        role: str | None = None,
        principal: str | None = None,
        collection: t.Union[uuid.UUID, str] | None = None,
        **kwargs: t.Any,
    ) -> None:
        _set_string_kwargs(
            self,
            DATA_TYPE=DATA_TYPE,
            role=role,
            principal=principal,
            collection=collection,
        )
        self.update(kwargs)


class NodeDocument(dict):
    """
    Convenience class for constructing a Node document to use as the `data`
    parameter to :meth:`node_update`

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.
    """

    def __init__(
        self,
        # For backwards compatibility
        DATA_TYPE: str = "node#1.0.0",
        **kwargs: t.Any,
    ) -> None:
        if "data_interface6" in kwargs:
            DATA_TYPE = "node#1.2.0"
        elif "data_interface" in kwargs:
            DATA_TYPE = "node#1.1.0"
        self["DATA_TYPE"] = DATA_TYPE
        self.update(kwargs)


class SharingPolicyDocument(dict):
    """
    Convenience class for constructing a SharingPolicy document to use as the
    `data` parameter to :meth:`create_sharing_policy`

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.
    """

    def __init__(
        self,
        DATA_TYPE: str = "sharing_policy#1.0.0",
        users: abc.Iterable[str] | None = None,
        read: abc.Iterable[str] | None = None,
        read_write: abc.Iterable[str] | None = None,
        none: abc.Iterable[str] | None = None,
        **kwargs: t.Any,
    ) -> None:
        if users:
            self["users"] = users
        self["DATA_TYPE"] = DATA_TYPE
        self["sharing_restrict_paths"] = {
            "DATA_TYPE": "path_restrictions#1.0.0",
            "read": read if read else [],
            "read_write": read_write if read_write else [],
            "none": none if none else [],
        }

        self.update(kwargs)


class OwnerStringDocument(dict):
    """
    Convenience class for constructing a o document to use as the
    `data` parameter to :meth:`create_sharing_policy`

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.
    ``identity_id`` (*uuid*)
      New advertised owner id
    """

    def __init__(
        self,
        DATA_TYPE: str = "owner_string#1.0.0",
        identity_id: str | None = None,
        **kwargs: t.Any,
    ) -> None:
        self["DATA_TYPE"] = DATA_TYPE
        self["identity_id"] = identity_id

        self.update(kwargs)


class UserCredentialDocument(dict):
    """
    Convenience class for constructing a User Credential document
    to use as the `data` parameter to :meth:`create_user_credential` or
    :meth:`update_user_credential`

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``identity_id`` (*string*)
      Globus Auth identity to who acts as the owner of this credential.

    ``connector_id`` (*string*)
      UUID of the connector type that this credential interacts with.

    ``username`` (*string*)
      Connector-specific username that this credential is associated with. If
      the connector supports identity mapping, this matches the result of the
      mapping applied to identity_id.

    ``display_name`` (*string*)
      Display name of the credential.

    ``storage_gateway_id`` (*string*)
      UUID of the storage gateway that this credential is associated with.

    ``home_directory`` (*string*)
      The home directory of this account associated with this credential.

    ``invalid`` (*bool*)
      Flag indicating that this credential is no longer valid.

    ``provisioned`` (*bool*)
      Flag indicating that this credential has been fully provisioned. If this
      is false and the invalid property is true, then the credential was
      created during login and patching it to add the missing data should be
      presented to the user as initializing the credential.

    ``policies`` (*dict*)
      Connector specific storage policies. It is recommended that
      you use one of the Policy helper classes (e.g. `S3UserCredentialPolicies`
      if you are using the S3 connector) to create these.
    """

    def __init__(
        self,
        DATA_TYPE: str = "user_credential#1.0.0",
        identity_id: str | None = None,
        connector_id: str | None = None,
        username: str | None = None,
        display_name: str | None = None,
        storage_gateway_id: str | None = None,
        home_directory: str | None = None,
        invalid: bool | None = None,
        provisioned: bool | None = None,
        policies: abc.Mapping | None = None,
        **kwargs: t.Any,
    ) -> None:
        _set_string_kwargs(
            self,
            DATA_TYPE=DATA_TYPE,
            identity_id=identity_id,
            connector_id=connector_id,
            username=username,
            display_name=display_name,
            storage_gateway_id=storage_gateway_id,
            home_directory=home_directory,
        )
        _set_raw_kwargs(
            self,
            invalid=invalid,
            provisioned=provisioned,
            policies=policies,
        )

        self.update(kwargs)


class S3UserCredentialPolicies(dict):
    """
    Convenience class for constructing a S3 User Credential Policy
    document to use as the `policies` parameter when creating a
    UserCredentialDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``s3_key_id`` (*strings*)
      Access Key ID to use with the S3 API to access your buckets and objects.

    ``s3_secret_key`` (*strings*)
      Secret Key to use with the S3 API to access your buckets and objects

    ``s3_requester_pays`` (*bool*)
      Acknowledgement that the account will pay for access to the S3 resource
    """

    def __init__(
        self,
        DATA_TYPE: str | None = None,
        s3_key_id: str | None = None,
        s3_secret_key: str | None = None,
        s3_requester_pays: bool | None = None,
        **kwargs: t.Any,
    ) -> None:
        if DATA_TYPE is None:
            if s3_requester_pays is not None:
                DATA_TYPE = "s3_user_credential_policies#1.1.0"
            else:
                DATA_TYPE = "s3_user_credential_policies#1.0.0"

        _set_bool_kwargs(self, s3_requester_pays=s3_requester_pays)
        _set_string_kwargs(
            self,
            DATA_TYPE=DATA_TYPE,
            s3_key_id=s3_key_id,
            s3_secret_key=s3_secret_key,
        )
        self.update(kwargs)


class ActiveScaleUserCredentialPolicies(dict):
    """
    Convenience class for constructing an ActiveScale User Credential Policy
    document to use as the `policies` parameter when creating a
    UserCredentialDocument

    **Parameters**

    ``DATA_TYPE`` (*string*)
      Versioned document type.

    ``s3_key_id`` (*strings*)
      Access Key ID to use with the S3 API to access your buckets and objects.

    ``s3_secret_key`` (*strings*)
      Secret Key to use with the S3 API to access your buckets and objects
    """

    def __init__(
        self,
        DATA_TYPE: str = "activescale_user_credential_policies#1.0.0",
        s3_key_id: str | None = None,
        s3_secret_key: str | None = None,
        **kwargs: t.Any,
    ) -> None:
        _set_string_kwargs(
            self, DATA_TYPE=DATA_TYPE, s3_key_id=s3_key_id, s3_secret_key=s3_secret_key
        )
        self.update(kwargs)


class BatchDocument(dict):
    def __init__(
        self,
        ids: abc.Iterable[t.Union[str, uuid.UUID]],
    ) -> None:
        self["DATA_TYPE"] = "batch#1.0.0"
        self["ids"] = [str(i) for i in list(ids)]


class AuthClientDocument(dict):
    def __init__(
        self,
        project: uuid.UUID,
        name: str,
        public_client: bool,
        client_type: str,
    ) -> None:
        self["client"] = {}
        _set_string_kwargs(
            self["client"],
            project=project,
            name=name,
            client_type=client_type,
        )
        _set_bool_kwargs(self["client"], public_client=public_client)


class AuthScopeDocument(dict):
    def __init__(
        self,
        name: str,
        description: str,
        scope_suffix: str,
        advertised: bool,
        allow_refresh_tokens: bool,
        dependent_scopes: abc.Sequence[abc.Mapping[str, t.Any]],
    ) -> None:
        self["scope"] = {}
        _set_string_kwargs(
            self["scope"],
            name=name,
            description=description,
            scope_suffix=scope_suffix,
        )
        _set_bool_kwargs(
            self["scope"],
            advertised=advertised,
            allow_refresh_tokens=allow_refresh_tokens,
        )

        self["scope"]["dependent_scopes"] = []

        for index in range(len(dependent_scopes)):
            self["scope"]["dependent_scopes"].append({})
            dep_scope = dependent_scopes[index]
            _set_string_kwargs(
                self["scope"]["dependent_scopes"][index], scope=dep_scope["scope"]
            )
            _set_bool_kwargs(
                self["scope"]["dependent_scopes"][index],
                optional=dep_scope["optional"],
                requires_refresh_token=dep_scope["requires_refresh_token"],
            )


class AuthCredentialDocument(dict):
    def __init__(self, name: str) -> None:
        self["credential"] = {}
        _set_string_kwargs(self["credential"], name=name)
