from typing import Optional

from flask import current_app, jsonify

from globus.manager.api.blueprints.decorators import (
    Base62Marker,
    Paginator,
    paginated,
    result_data,
)
from globus.manager.api.models.result import Result
from globus.manager.api.schemas.api import data_type_schema as data_types
from globus.manager.api.utils.info import info
from globus.manager.version import __api_version__ as api_version

from .openapi import OpenAPIBlueprint

blueprint = OpenAPIBlueprint(
    "info_blueprint",
    __name__,
    tag="Versioning",
    description=f"""
    ### API Versioning
    The GCS Manager API uses [semantic versioning](https://semver.org/) for
    its API versioning. The version of the API described in this document
    is {api_version}.

    ### Connectors
    Each Connector providing access to a type of storage will have its
    own semantic version. This version describes the API specific
    to that connector, including its input and output document formats
    and any additional URL routes that it provides.

    ### API Document Versions
    Each document in the Globus Connect Server Manager API includes
    a `DATA_TYPE` property that indicates the type and version of the
    document. When a document with a given `DATA_TYPE` is used in a
    document, the document must validate against the schema defined
    for that document type. Note that extensions may be allowed based
    on the definition of that type.

    ### Deprecations
    When a part of the API is to be deprecated, the minor version of the
    API is increased.  When a part of the API is removed, the major version
    of the API is incremented.
    When a deprecated API is used, the response message will include a
    deprecation warning.

    """,
    menu_weight=2,
    url_prefix="/api/",
)


@blueprint.route(  # type: ignore
    "/info",
    methods=["GET"],
    summary="Get GCS service information",
)
@result_data(
    schemas=[data_types.Connector, data_types.Info],
    description="Get info response",
)
@paginated
def info_get(*, page_size: int, marker: Optional[Paginator]) -> Result:
    """
    Returns information about the GCS Manager service for this endpoint, as
    well as additional features such as connectors that it provides as
    extensions to the API defined in this document.

    This operation can be performed without an `Authorization` header.
    """
    next_page_marker = None
    data = []

    if not marker:
        data.append(data_types.Info.dump(info()))

    connectors = [
        current_app.config["connectors"][k]
        for k in sorted(current_app.config["connectors"].keys())
        if k >= (marker.marker if marker else "")
    ]

    for conn in connectors:
        if len(data) == page_size:
            next_page_marker = Base62Marker(marker_data=conn.CONNECTOR_ID)
            break

        connector_info = conn.model.connector

        if connector_info.enabled:
            data.append(data_types.Connector.dump(connector_info))

    res = data_types.Result.dump(Result(data=data, marker=next_page_marker))

    return jsonify(res)
