from __future__ import annotations

import enum
import uuid
from typing import Any, Dict, List, Optional

from flask import Response, current_app, g, jsonify, request, url_for
from sqlalchemy.sql.expression import false

import globus_sdk

import globus.manager.api.errors as errors
import globus.manager.api.models as models
from globus.manager.api.authentication import AuthenticationState
from globus.manager.api.authentication.decorators import authenticated
from globus.manager.api.authorization import (
    EndpointRoleAuthorizer,
    IdentitySet,
    PublicAuthorizer,
    RoleSpec,
    StorageGatewayAuthorizer,
)
from globus.manager.api.blueprints.decorators import (
    Base62Marker,
    Paginator,
    db_read,
    document_authorization,
    enum_list_converter,
    expects_json,
    paginated,
    query_parameter,
    requires_credential_owner,
    requires_one_role,
    result_data,
    server_timing,
)
from globus.manager.api.blueprints.utils import (
    check_subscription,
    create_object,
    patch_object,
    put_object,
    update_object,
)
from globus.manager.api.models import Result, db
from globus.manager.api.schemas.api import data_type_schema as data_types
from globus.manager.api.utils.log import log_auth_exception
from globus.manager.constants.roles import RoleType

from .openapi import (
    OpenAPIBlueprint,
    RoleAuthorizer as OpenAPIRoleAuthorizer,
    StorageGatewayAuthorizer as OpenAPIStorageGatewayAuthorizer,
)

blueprint = OpenAPIBlueprint(
    "user_credentials",
    __name__,
    tag="User Credentials",
    description="""
    The User Credential API allows users to manage credentials needed to access
    storage gateways. The credentials may take different forms, based on the
    connector being used.
    """,
    menu_weight=8,
    url_prefix="/api",
)


class UserCredentialGetInclude(enum.Enum):
    all = "all"


@blueprint.route(
    "/user_credentials",
    methods=["GET"],
    summary="List user credentials",
    operationId="listUserCredentials",
)
@result_data(
    schemas=[data_types.UserCredential],
    description="List user credential response",
    max_items=1,
)
@query_parameter(
    name_spec="<uuid:storage_gateway>",
    description="ID of the Storage Gateway",
)
@query_parameter(
    name_spec="include",
    description="""
    Credentials to include. If set to **all**, then -
    credentials owned by other users are returned if the storage gateway allows
    admin_managed_credentials and the caller has an endpoint:administrator or
    endpoint:owner role.
    """,
    converter=enum_list_converter(UserCredentialGetInclude),
    default=[],
)
@document_authorization(
    authorization=OpenAPIStorageGatewayAuthorizer(scopes=["{storage_gateway_id}"])
)
@document_authorization(
    authorization=OpenAPIRoleAuthorizer(
        roles=[
            RoleType.endpoint_owner.value,
            RoleType.endpoint_administrator.value,
        ]
    )
)
@paginated
@server_timing
@authenticated
def user_credentials_get(
    *,
    page_size: int,
    marker: Optional[Paginator],
    storage_gateway: uuid.UUID,
    include: List[UserCredentialGetInclude],
) -> Response:
    """
    The caller may only retrieve User Credentials which were created by an
    identity in the caller's identity set unless they are an administrator. If
    the endpoint is not public, returns ForbiddenError if the user has no
    credentials
    """
    endpoint_is_public = PublicAuthorizer().authorize_identities(g.auth_state)

    if storage_gateway:
        sg = models.StorageGateway.query.filter(
            models.StorageGateway.id == str(storage_gateway),
            ~models.StorageGateway.deleted,  # type: ignore
        ).one_or_none()
        if not sg:
            raise errors.NotFoundError()
        authorizer = StorageGatewayAuthorizer(sg, enforce_authentication_timeout=False)
        authorized_identities = authorizer.authorize_identities(g.auth_state)
    else:
        authorized_identities = IdentitySet()
        for i in g.auth_state.identity_set_data:
            authorized_identities.append(i)

    # include=all will return all credentials that aren't deleted from storage
    # gateways that have admin_managed_credentials=True as well as all of the
    # user's credentials, if the caller has an endpoint:administrator or
    # endpoint:owner role
    include_all = False

    if UserCredentialGetInclude.all in include:
        try:
            requires_one_role(
                role_names=[
                    RoleType.endpoint_administrator.value,
                    RoleType.endpoint_owner.value,
                ],
            )(lambda: True)()
            include_all = True
        except errors.ForbiddenError as e:
            current_app.logger.debug(
                f"Role authorization for user_credentials_post() failed with {e}"
            )

    # List of all credentials which are not deleted
    query = models.UserCredential.query.filter(
        models.UserCredential.deleted == false(),
    )

    if not include_all:
        query = query.filter(
            models.UserCredential.identity_id.in_(  # type: ignore
                [i["id"] for i in authorized_identities]
            )
        )
    if storage_gateway:
        query = query.filter_by(storage_gateway_id=str(storage_gateway))

    if marker:
        query = query.filter(models.ConfigurationObject.id >= marker.marker)

    query = query.order_by(models.ConfigurationObject.id).limit(page_size + 1)

    g.server_timing.start(db_read)
    query_results = []
    returned_rows = 0
    last_id = None

    # storage_gateway_id: (identity_id, local_user)
    account_mappings: dict[str, list[tuple[str, str]]] = {}
    for credential in query:
        last_id = credential.id
        returned_rows += 1

        if returned_rows > page_size:
            break

        if credential.is_deleted:
            continue

        if not (include_all and credential.storage_gateway.admin_managed_credentials):
            if credential.storage_gateway_id not in account_mappings:
                account_mappings[credential.storage_gateway_id] = [
                    (a.identity_id, a.username)
                    for a in credential.storage_gateway.map_identities(
                        authorized_identities
                    )
                ]
            sg_accounts = account_mappings[credential.storage_gateway_id]
            if (credential.identity_id, credential.username) in sg_accounts:
                query_results.append(credential)
        else:
            query_results.append(credential)
    g.server_timing.finish(db_read)

    # If the caller has an endpoint role, they can see all collections
    caller_has_endpoint_role = EndpointRoleAuthorizer.authorize_identities(g.auth_state)

    if query_results == [] and not endpoint_is_public and not caller_has_endpoint_role:
        raise errors.MissingRequiredRoleError(
            roles=[
                RoleSpec("endpoint:owner"),
                RoleSpec("endpoint:administrator"),
                RoleSpec("endpoint:activity_manager"),
                RoleSpec("endpoint:activity_monitor"),
            ]
        )

    results = []
    for uc in query_results:
        results.append(data_types.UserCredential.dump(uc))

    if returned_rows > page_size:
        next_page_token = Base62Marker(marker_data=last_id)
    else:
        next_page_token = None

    res = data_types.Result.dump(Result(data=results, marker=next_page_token))

    resp = jsonify(res)
    resp.status_code = res["http_response_code"]
    resp.headers["Server-Timing"] = g.server_timing.as_string()

    return resp


def get_user_credential_id_from_path() -> Optional[models.UserCredential]:
    """
    Resolves the user_credential_id view parameter to a UserCredential if
    a UserCredential with that ID is in the database.
    """
    user_credential_id = request.view_args.get("user_credential_id")

    assert user_credential_id is not None

    user_credential = None
    if user_credential_id:
        user_credential = models.UserCredential.query.filter(
            models.UserCredential.id == str(user_credential_id)
        ).one_or_none()
        if user_credential and user_credential.is_deleted:
            user_credential = None

    g.user_credential = user_credential
    return user_credential


@blueprint.route(
    "/user_credentials/<uuid:user_credential_id>",
    methods=["GET"],
    summary="Get a user credential",
    path_parameters={"user_credential_id": "Id of the user credential"},
    operationId="getUserCredential",
)
@result_data(
    schemas=[data_types.UserCredential],
    description="Get user credential response",
    max_items=1,
)
@authenticated
@requires_credential_owner(
    credential_selector=get_user_credential_id_from_path,
    allow_admin_if_configured=True,
)
def user_credentials_get_by_id(user_credential_id: uuid.UUID) -> Response:
    """
    Get a user credential.

    The caller must have the identity_id of the user credential
    in its identity set.
    """
    uc = g.user_credential

    result = data_types.UserCredential.dump(uc)

    res = data_types.Result.dump(Result(data=[result]))

    resp = jsonify(res)
    resp.status_code = res["http_response_code"]

    return resp


def get_storage_gateway_from_credential_doc() -> Optional[models.StorageGateway]:
    storage_gateway_id = g.validated_data.get("storage_gateway_id")
    if storage_gateway_id:
        storage_gateway = models.StorageGateway.query.filter(
            models.StorageGateway.id == str(storage_gateway_id)
        ).one_or_none()
        g.storage_gateway = storage_gateway
        return storage_gateway
    return None


@blueprint.route(
    "/user_credentials/<uuid:user_credential_id>",
    methods=["PATCH"],
    summary="Update a user credential",
    path_parameters={"user_credential_id": "Id of the user credential"},
    operationId="patchUserCredential",
)
@result_data(
    schemas=[data_types.UserCredential],
    description="Update user credential response",
    max_items=1,
)
@authenticated
@expects_json(
    schema=data_types.UserCredential,
    patch_schema=True,
    description="Changes to the user credential definition",
    required=True,
)
@requires_credential_owner(
    credential_selector=get_user_credential_id_from_path,
    allow_admin_if_configured=True,
)
def user_credentials_patch(user_credential_id: uuid.UUID) -> Response:
    """
    Update a user credential on a storage gateway. This is required
    for some connectors that require a local user name or other
    credential information to access the storage system. See the
    connector-specific documentation for details on what is needed.

    The caller must have the identity_id of the user credential
    in its identity set.
    """
    uc = g.user_credential

    with current_app.appsync_client.pause_notifications():
        if uc.is_deleted:
            current_app.logger.error("PATCH %s: already deleted", str(uc.id))
            raise errors.NotFoundError()
        elif g.validated_data.get("deleted"):
            raise errors.ValidationError("Cannot change deleted via PATCH")

        patch_object(uc, g.validated_data)
        res = data_types.Result.dump(
            Result(
                message="Updated User Credential {}".format(str(uc.id)),
                data=[data_types.UserCredential.dump(uc)],
            )
        )

    resp = jsonify(res)
    resp.status_code = res["http_response_code"]
    g.updated_objects.append(uc.id)

    return resp


@blueprint.route(
    "/user_credentials/<uuid:user_credential_id>",
    methods=["PUT"],
    summary="Update a user credential",
    path_parameters={"user_credential_id": "Id of the user credential"},
    operationId="putUserCredential",
)
@result_data(
    schemas=[data_types.UserCredential],
    description="Update user credential response",
    max_items=1,
)
@authenticated
@expects_json(
    schema=data_types.UserCredential,
    description="Updated user credential definition",
    required=True,
)
@requires_credential_owner(
    credential_selector=get_user_credential_id_from_path,
    allow_admin_if_configured=True,
)
def user_credentials_put(user_credential_id: uuid.UUID) -> Response:
    """
    Update a user credential on a storage gateway. This is required
    for some connectors that require a local user name or other
    credential information to access the storage system. See the
    connector-specific documentation for details on what is needed.

    The caller must have the identity_id of the user credential
    in its identity set.
    """
    uc = g.user_credential

    with current_app.appsync_client.pause_notifications():
        if uc.is_deleted:
            current_app.logger.error("PUT %s: already deleted", str(uc.id))
            raise errors.NotFoundError()
        elif g.validated_data.get("deleted", False) != g.user_credential.deleted:
            raise errors.ValidationError("Cannot change deleted via PUT")

        put_object(uc, g.validated_data)
        res = data_types.Result.dump(
            Result(
                message="Updated User Credential {}".format(str(uc.id)),
                data=[data_types.UserCredential.dump(uc)],
            )
        )

    resp = jsonify(res)
    resp.status_code = res["http_response_code"]
    g.updated_objects.append(uc.id)

    return resp


@blueprint.route(
    "/user_credentials",
    methods=["POST"],
    summary="Create a user credential",
    operationId="postUserCredential",
)
@result_data(
    schemas=[data_types.UserCredential],
    description="Create user credential response",
    http_status=201,
    max_items=1,
)
@authenticated
@expects_json(
    schema=data_types.UserCredential,
    description="New user credential definition",
    required=True,
)
def user_credentials_post() -> Response:
    """
    Create a user credential on a storage gateway. This is required
    for some connectors that require a local user name or other
    credential information to access the storage system. See the
    connector-specific documentation for details on what is needed.

    The caller is authorized based on the StorageGateway identity
    policies, so users with no assigned roles on the endpoint may
    be permitted to access this operation.
    """
    g.storage_gateway = get_storage_gateway_from_credential_doc()
    if g.storage_gateway is None:
        raise errors.BadRequestError()

    assert g.validated_data is not None
    check_subscription(
        g.storage_gateway.connector,
        high_assurance=g.storage_gateway.high_assurance,
    )

    has_role = False
    meets_gateway_policy = False

    if g.storage_gateway.admin_managed_credentials:
        # Only permit role authorization if allowed by policy
        try:
            has_role = requires_one_role(
                storage_gateway_selector=lambda: g.storage_gateway,
                role_names=[
                    RoleType.endpoint_administrator.value,
                    RoleType.endpoint_owner.value,
                ],
            )(lambda: True)()
        except errors.ForbiddenError as e:
            current_app.logger.debug(
                f"Role authorization for user_credentials_post() failed with {e}"
            )

    # Check if ids in the caller's ID set are allowed by storage gateway
    # policies.
    storage_gateway_authorizer = StorageGatewayAuthorizer(g.storage_gateway)
    sg_auth_result = storage_gateway_authorizer.authorize_identities(g.auth_state)
    if sg_auth_result:
        meets_gateway_policy = True
    elif not has_role:
        detail = None
        if sg_auth_result.error_details:
            detail = sg_auth_result.error_details[0]
        raise errors.AuthorizationError(detail=detail)

    assert has_role or meets_gateway_policy

    # identities which are authorized by storage gateway policy, indexed by the id
    authorized_identities: Dict[str, Dict[str, Any]] = {}

    if meets_gateway_policy:
        authorized_identities = {i["id"]: i for i in sg_auth_result}

    if "identity_id" not in g.validated_data:
        # Infer identity_id if not present. This can only be done if the caller
        # has a single identity in their identity set which meets the storage
        # gateway policy.
        if not meets_gateway_policy:
            raise errors.ValidationError(
                error_description="Missing identity_id property"
            )

        if len(authorized_identities) == 1:
            g.validated_data["identity_id"] = list(authorized_identities.keys())[0]
        else:
            raise errors.ValidationError(
                "Missing identity_id property, and you have more than one "
                "authorized Globus account. Set the identity_id property on the "
                "input document to select which of your identities to use."
            )
    assert g.validated_data["identity_id"] is not None
    requested_identity_id: str = str(g.validated_data["identity_id"])

    # identity document for the requested_identity_id value.
    authorized_identity: Optional[Dict[str, Any]] = None

    # The caller must either have the requested identity in their identity
    # set or have the endpoint role. In the latter case, we'll use the auth
    # client to resolve the identity_id to an identity dict and then validate
    # that identity against the storage gateway policies
    if requested_identity_id in authorized_identities.keys() and meets_gateway_policy:
        authorized_identity = authorized_identities[requested_identity_id]
    elif has_role:
        try:
            # Determine if the requested identity meets storage gateway policy
            # if we are authorizing because of a role. First we look up the
            # requested identity in auth and then pass it to the
            # storage_gateway_authorizer using a synthesized
            resolved_identities: List[Dict[str, Any]] = [
                i
                for i in current_app.auth_client.get_identities(
                    ids=requested_identity_id
                )
                if i["id"] == requested_identity_id
            ]
            auth_state = AuthenticationState(identity_set_data=resolved_identities)
            # New authorizer that does not enforce timeouts, since we don't
            # have any authentication information related to the identity
            # authorized by the role
            storage_gateway_authorizer = StorageGatewayAuthorizer(
                g.storage_gateway, enforce_authentication_timeout=False
            )
            sg_auth_result = storage_gateway_authorizer.authorize_identities(auth_state)
            if not sg_auth_result:
                detail = None
                if sg_auth_result.error_details:
                    detail = sg_auth_result.error_details[0]
                raise errors.AuthorizationError(detail=detail)
            authorized_identity = sg_auth_result[0]
        except globus_sdk.GlobusAPIError as e:
            log_auth_exception(e)
    else:
        raise errors.IdNotInIdentitySetError(id=requested_identity_id)

    if not authorized_identity:
        raise errors.BadRequestError(f"Invalid identity_id {requested_identity_id}")

    mappings = g.storage_gateway.map_identities(
        IdentitySet(identity_set=[authorized_identity])
    )

    if not mappings:
        raise errors.InvalidUserError()

    mapped_to = [mapping.username for mapping in mappings]

    requested_username = g.validated_data.get("username")
    if requested_username is None:
        g.validated_data["username"] = mapped_to[0]

    if requested_username and requested_username not in mapped_to:
        raise errors.InvalidUserError(usernames=mapped_to)
    if not g.storage_gateway.check_user_allowed(g.validated_data["username"]):
        raise errors.InvalidUserError(usernames=[g.validated_data["username"]])

    with current_app.appsync_client.pause_notifications():
        try:
            cls = models.UserCredential.polymorphic_class(
                g.storage_gateway.connector_id
            )
            old_uc = cls.query.filter(
                cls.storage_gateway_id == g.storage_gateway.id,
                cls.username == g.validated_data["username"],
                cls.identity_id == requested_identity_id,
            ).one_or_none()

            if old_uc:
                if not old_uc.deleted:
                    raise errors.ResourceExistsError(resource_ids=[old_uc.id])
            elif not old_uc:
                uc = cls(doc=g.validated_data, storage_gateway=g.storage_gateway)
        except ValueError as e:
            current_app.logger.error("Error creating using credential: {}".format(e))
            raise errors.ValidationError()

        if old_uc:
            old_uc.deleted = False
            put_object(old_uc, g.validated_data)
            uc = old_uc
        else:
            create_object(uc)

        credential_json = data_types.UserCredential.dump(uc)

        res = data_types.Result.dump(
            Result(
                http_response_code=201,
                message="Created User Credential {}".format(uc.id),
                data=[credential_json],
            )
        )

    response = jsonify(res)
    response.status_code = res["http_response_code"]
    response.headers["Location"] = url_for(
        ".user_credentials_get_by_id",
        user_credential_id=uc.id,
        _external=True,
        _scheme="https",
    )
    g.created_objects.append(uc.id)

    return response


@blueprint.route(
    "/user_credentials/<uuid:user_credential_id>",
    methods=["DELETE"],
    summary="Delete a user credential",
    path_parameters={"user_credential_id": "Id of the user credential"},
    operationId="deleteUserCredential",
)
@result_data(
    schemas=[],
    description="Delete user credential response",
)
@authenticated
@requires_credential_owner(
    credential_selector=get_user_credential_id_from_path, allow_admin_if_configured=True
)
def user_credentials_delete(user_credential_id: uuid.UUID) -> Response:
    """
    Delete a user credential.

    The caller must have the identity_id of the user credential in its identity
    set.
    """
    uc = g.user_credential

    with current_app.appsync_client.pause_notifications():
        if uc.is_deleted:
            current_app.logger.error("DELETE %s: already deleted", str(uc.id))
            raise errors.NotFoundError()

        existing_collections = [
            uc.id
            for c in db.session.query(models.Collection.id)
            .filter(models.Collection.user_credential_id == uc.id)
            .all()
        ]

        if existing_collections:
            current_app.logger.info(
                "DELETE %s: in use by %d Collection(s)",
                uc.id,
                len(existing_collections),
            )

        def delete_cred(credential: models.UserCredential) -> None:
            credential.delete()

        update_object(uc, delete_cred)

        res = data_types.Result.dump(
            Result(
                message="Deleted User Credential {}".format(str(user_credential_id)),
                data=[],
            )
        )

    resp = jsonify(res)
    resp.status_code = res["http_response_code"]
    g.deleted_objects.append(uc.id)

    return resp
