from __future__ import annotations

import typing as t

import globus.manager.api.errors as errors
import globus.manager.api.models as m
from globus.manager.constants import CollectionType, ConfigurationObjectTag

from .collection import Collection


class GoogleCloudStorageCollection(Collection):
    POLICY_DATA_TYPE = "google_cloud_storage_collection_policies#1.0.0"
    POLICY_DATA_TYPES = [POLICY_DATA_TYPE]

    CONNECTOR_POLICIES = {POLICY_DATA_TYPE: ["project"]}

    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.COLLECTION.name
            + "56366b96-ac98-11e9-abac-9cb6d0d9fd63"
        )
    }

    def __init__(
        self,
        appsync_data=None,
        doc=None,
        storage_gateway=None,
        mapped_collection=None,
        user_credential=None,
        endpoint=None,
    ):
        if doc:
            policies = doc.get("policies")
            if not policies:
                policies = {
                    "DATA_TYPE": self.POLICY_DATA_TYPE,
                }

            project = policies.get("project")
            collection_type = CollectionType(doc.get("collection_type"))

            if collection_type == CollectionType.guest:
                if project is None:
                    project = getattr(mapped_collection, "project", None)
                if not user_credential:
                    raise ValueError("user_credential")

                # Raises exception if not possible
                user_credential.can_access_project(project)
            elif project is None:
                # If the storage gateway defines only one project, use it as
                # the default if none is specified
                allowed_projects = getattr(storage_gateway, "projects", [])
                if len(allowed_projects) == 1:
                    project = allowed_projects[0]
                if project is None and len(allowed_projects) > 1:
                    raise ValueError(
                        "The policies document for this collection is missing the "
                        "project property. This is required to create a Google "
                        "Cloud Storage collection unless the storage gateway "
                        "allows all projects or is restricted to a single project. "
                        "See the documentation at https://docs.globus.org/"
                        "premium-storage-connectors/v5.4/google-cloud-storage/"
                        "#collection_policies for more details."
                    )

            # Raises an exception if the project property doesn't meet storage
            # gateway policies
            self._check_project(project, storage_gateway, mapped_collection)

            policies = {**policies, **{"project": project}}
            doc = {**doc, "policies": {**policies}}

        super().__init__(
            appsync_data=appsync_data,
            doc=doc,
            storage_gateway=storage_gateway,
            mapped_collection=mapped_collection,
            user_credential=user_credential,
            endpoint=endpoint,
        )

    @property
    def connector_id(self):
        """
        UUID of the Google Cloud Storage connector
        """
        return "56366b96-ac98-11e9-abac-9cb6d0d9fd63"

    @connector_id.setter
    def connector_id(self, value):
        pass

    @property
    def project(self):
        """
        Project associated with this collection. All users of this collection
        must have Google credentials which are authorized to use this account,
        and all access to Google Cloud Storage buckets will be charged to this
        account.
        """
        if self.is_guest_collection:
            return self.mapped_collection.project
        else:
            return self.get_policy("project")

    @project.setter
    def project(self, value):
        if self.is_mapped_collection:
            self.set_policy("project", value)

    @project.deleter
    def project(self):
        self.delete_policy("project")

    def put(self, **doc):
        """
        Overload of put() that checks policies for validity. Check here and in
        check_user_credential instead of in policy setters so that object loads
        aren't broken when a collection is in an inconsistent state.
        """
        policies = doc.get("policies", {})
        if policies and policies.get("DATA_TYPE") != self.POLICY_DATA_TYPE:
            raise ValueError("Unsupported DATA_TYPE")

        project = policies.get("project")

        # guests default to mapped project
        if not project and not self.is_mapped_collection:
            project = getattr(self.mapped_collection, "project", None)

        self._check_project(
            project,
            self.storage_gateway,
            self.mapped_collection if self.is_guest_collection else None,
        )

        super().put(**doc)

    def patch(self, **doc):
        """
        Overload of patch() that checks policies for validity. Check here and in
        check_user_credential instead of in policy setters so that object loads
        aren't broken when a collection is in an inconsistent state.
        """
        policies = doc.get("policies", {})
        if policies and policies.get("DATA_TYPE") != self.POLICY_DATA_TYPE:
            raise ValueError("Unsupported DATA_TYPE")

        if "project" in policies:
            project = policies.get("project")
            self._check_project(
                project,
                self.storage_gateway,
                self.mapped_collection if self.is_guest_collection else None,
            )

        super().patch(**doc)

    def _check_project(
        self,
        project: str,
        storage_gateway: m.StorageGateway,
        mapped_collection: t.Optional[m.Collection] = None,
    ) -> None:
        """
        Check that the project requested by the user creating or
        updating a collection makes sense. This will check that the
        """
        allowed_projects = getattr(storage_gateway, "projects", None)
        # collection project is optional only if there is no sg project list
        if project or allowed_projects:
            if not project and allowed_projects:
                raise ValueError(
                    "A collection project {} is required by the "
                    "policies of storage gateway {}".format(project, storage_gateway.id)
                )
            if allowed_projects and project not in allowed_projects:
                raise ValueError(
                    "Project {} is not allowed by the "
                    "policies of storage gateway {}".format(project, storage_gateway.id)
                )
            if mapped_collection and project != getattr(
                mapped_collection, "project", None
            ):
                raise ValueError(
                    "Project {} is not the same as the value used by the "
                    "policies of mapped collection {}".format(
                        project, mapped_collection.id
                    )
                )

    def check_user_credential(self, credential):
        project = getattr(self, "project", None)
        try:
            self._check_project(
                project,
                self.storage_gateway,
                self.mapped_collection if self.is_guest_collection else None,
            )
        except ValueError as v:
            raise errors.ForbiddenError(str(v))
        credential.can_access_project(project)

    def __repr__(self):
        return "<GoogleCloudStorageCollection {}>".format(self.id)
