from __future__ import annotations

from flask import current_app

from globus.manager.constants import ConfigurationObjectTag

from .connector import Connector
from .decorators import checks_policy_property
from .storage_gateway import StorageGateway


class GoogleDriveStorageGateway(StorageGateway):
    CONNECTOR_ID = "976cf0cf-78c3-4aab-82d2-7c16adbcc281"
    POLICY_DATA_TYPE = "google_drive_storage_policies#1.1.0"
    CONNECTOR_POLICIES = {
        "google_drive_storage_policies#1.0.0": [
            "client_id",
            "secret",
            "user_api_rate_quota",
        ]
    }
    CONNECTOR_POLICIES["google_drive_storage_policies#1.1.0"] = CONNECTOR_POLICIES[
        "google_drive_storage_policies#1.0.0"
    ] + ["allow_any_account"]

    POLICY_DATA_TYPES = list(CONNECTOR_POLICIES.keys())
    connector_policy_names = CONNECTOR_POLICIES[POLICY_DATA_TYPE]

    __mapper_args__ = {
        "polymorphic_identity": (
            ConfigurationObjectTag.STORAGE_GATEWAY.name + CONNECTOR_ID
        )
    }

    connector = Connector(
        id=CONNECTOR_ID, display_name="Google Drive", can_post_credential=False
    )

    client_id = StorageGateway.connector_property("client_id", required=True)

    @checks_policy_property(prop="client_id", required=True)
    def check_client_id(self, value):
        if not isinstance(value, str):
            raise ValueError()

    secret = StorageGateway.connector_property("secret", required=True)

    @checks_policy_property(prop="secret", required=True)
    def check_secret(self, value):
        if not isinstance(value, str):
            raise ValueError()

    user_api_rate_quota = StorageGateway.connector_property("user_api_rate_quota")

    @checks_policy_property(prop="user_api_rate_quota")
    def check_user_api_rate_quota(self, value: int):
        if value is not None and not isinstance(value, int):
            raise ValueError("Invalid user_api_rate_quota")

    allow_any_account = StorageGateway.connector_property("allow_any_account")

    @checks_policy_property(prop="allow_any_account")
    def check_allow_any_account(self, value: bool):
        if not isinstance(value, bool):
            raise ValueError("Invalid allow_any_account")

    def check_user_allowed(self, username):
        if self.client_id is None:
            current_app.logger.warning(
                "storage_gateway=%s client_id is not configured", self
            )
            return False

        if self.secret is None:
            current_app.logger.warning(
                "storage_gateway=%s secret is not configured", self
            )
            return False

        return super().check_user_allowed(username)

    def default_mapping(self, identity_data):
        if len(self.allowed_domains) == 1:
            user, domain = identity_data["username"].rsplit("@", 1)
            if domain in self.allowed_domains:
                return identity_data["username"]

    @property
    def load_dsi_module(self):
        return "google_drive"

    @load_dsi_module.setter
    def load_dsi_module(self, value):
        if value is not None and value != self.load_dsi_module:
            raise ValueError("Invalid load_dsi_module")

    def home_directory(self, username):
        return "/My Drive"

    def delete(self):
        self.client_id = None
        self.secret = None
        super().delete()

    @property
    def auth_callback(self):
        """
        URL of the auth callback that must be registered on the Google API
        console for the application client_id in order to process Google
        credentials.
        """
        try:
            return current_app.config["connectors"][self.CONNECTOR_ID].callback_url
        except Exception:
            raise AttributeError(
                "This storage gateway does not have a value for the auth_callback property"
            )

    @property
    def process_user(self):
        return "gcsweb"

    @process_user.setter
    def process_user(self, value):
        if value is not None and value != self.process_user:
            raise ValueError("Invalid process_user")
