from __future__ import annotations

import datetime

import httplib2
import marshmallow as ma
from flask import current_app
from oauth2client import client as google_oauth

import globus.manager.api.errors as errors

from .decorators import checks_property
from .user_credential import UserCredential


class OIDCUserCredential(UserCredential):
    epoch = datetime.datetime.fromtimestamp(0, datetime.timezone.utc)

    @property
    def connector_properties(self) -> list[str]:
        return [
            "sub",
            "email",
            "access_token",
            "refresh_token",
            "token_expiry",
            "scopes",
        ]

    sub = UserCredential.connector_property("sub")
    email = UserCredential.connector_property("email")
    access_token = UserCredential.connector_property("access_token")
    refresh_token = UserCredential.connector_property("refresh_token")
    scopes = UserCredential.connector_property("scopes")

    @scopes.setter  # type: ignore
    def scopes(self, value: str | list[str] | None):
        if not value:
            self.policies["scopes"] = None
        elif isinstance(value, str):
            self.policies["scopes"] = value.split()
        else:
            self.policies["scopes"] = value

    @refresh_token.setter  # type: ignore
    def refresh_token(self, value: str):
        self.policies["refresh_token"] = value
        if not value and not self.uses_service_auth:
            self.invalid = True

    @refresh_token.deleter  # type: ignore
    def refresh_token(self):
        if "refresh_token" in self.policies:
            del self.policies["refresh_token"]
        if not self.uses_service_auth:
            self.invalid = True

    @property
    def token_expiry(self) -> datetime.datetime:
        value = getattr(self, "_token_expiry", None)
        if not value:
            if "token_expiry" in self.policies:
                value = self._token_expiry = ma.utils.from_iso_datetime(
                    self.policies["token_expiry"]
                )
            else:
                value = self.epoch
            self._token_expiry = value

        if not value:
            raise AttributeError("token_expiry")
        return value

    @token_expiry.setter
    def token_expiry(self, value: str | datetime.datetime | None):
        if value is None:
            self.policies["token_expiry"] = self.epoch
            self._token_expiry = self.epoch
            return
        elif isinstance(value, str):
            dt = ma.utils.from_iso_datetime(value)
        else:
            assert isinstance(value, datetime.datetime)
            dt = value

        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=datetime.timezone.utc)
        self._token_expiry = dt
        self.policies["token_expiry"] = ma.utils.isoformat(dt)

    @token_expiry.deleter
    def token_expiry(self):
        self._token_expiry = self.epoch
        if "token_expiry" in self.configuration_hash:
            self.policies["token_expiry"] = ma.utils.isoformat(self.epoch)

    def needs_refresh(self):
        """
        Return True if credential requires some refresh action to be performed
        before the credential can be used. Subclasses should implement this
        if necessary to check for expiring credentials.
        """
        if self.deleted:
            return False

        try:
            five_mins_before_expiration = self.token_expiry - datetime.timedelta(
                minutes=5
            )
            now = datetime.datetime.now(datetime.timezone.utc)

            return self.invalid or (now > five_mins_before_expiration)
        except AttributeError:
            return True

    def refresh(self):
        """
        Update an expired temporary credential. Subclasses should implement
        this if necessary to refresh credentials.
        """
        try:
            gcred = google_oauth.GoogleCredentials(
                access_token=None,
                client_id=self.storage_gateway.client_id,
                client_secret=self.storage_gateway.secret,
                refresh_token=self.refresh_token,
                token_expiry=None,
                token_uri=self.token_uri,
                user_agent=None,
            )
            gcred.refresh(httplib2.Http())
        except Exception as e:
            current_app.logger.warning("Credential refresh failure: %s", str(e))
            if not self.invalid and not self.uses_service_auth:
                self.invalid = True
                self.configuration_hash.changed()
            raise errors.InvalidCredentialError(user_credential_id=self.id)

        self.refresh_token = gcred.refresh_token
        self.access_token = gcred.access_token
        self.token_expiry = gcred.token_expiry

        self.configuration_hash.changed()

    def delete(self):
        if self.supports_revoke:
            try:
                token = self.refresh_token
            except AttributeError:
                token = None
            if token:
                gcred = google_oauth.GoogleCredentials(
                    access_token=None,
                    client_id=None,
                    client_secret=None,
                    refresh_token=token,
                    token_expiry=None,
                    token_uri=self.token_uri,
                    user_agent=None,
                )
                try:
                    gcred.revoke(httplib2.Http())
                except google_oauth.TokenRevokeError as e:
                    current_app.logger.debug("delete: revoke: %s: %s", self.id, str(e))
                finally:
                    self.refresh_token = None
                    self.access_token = None
        else:
            self.refresh_token = None
            self.access_token = None
        super().delete()

    @checks_property(prop="display_name")
    def check_display_name(self, value: str):
        if value and getattr(self, "display_name", None) != value:
            raise AttributeError()

    def patch(self, **doc):
        display_name = doc.pop("display_name", getattr(self, "display_name", None))
        doc["display_name"] = display_name
        policies = doc.get("policies", None)
        if policies:
            if policies.get("DATA_TYPE") != self.POLICY_DATA_TYPE:
                raise ValueError("Unsupported DATA_TYPE")
            for p in policies:
                if p != "DATA_TYPE" and getattr(self, p, None) != policies[p]:
                    raise ValueError("Cannot change {} via PATCH".format(p))

        if (
            getattr(self, "refresh_token", None) is None and not self.uses_service_auth
        ) or self.invalid:
            if not doc.get("invalid", True):
                raise ValueError("Cannot mark this credential as valid")
            doc["invalid"] = True
        elif "invalid" not in doc:
            doc["invalid"] = self.invalid

        super().patch(**doc)

    def put(self, **doc):
        # Users cannot change policies via put, must be through put_credentials,
        # since those come from the google tokens.
        display_name = doc.pop("display_name", getattr(self, "display_name", None))
        doc["display_name"] = display_name

        policies = doc.get("policies", None)

        if policies and not self.uses_service_auth:
            if policies.get("DATA_TYPE") != self.POLICY_DATA_TYPE:
                raise ValueError("Unsupported DATA_TYPE")
            for p in policies:
                if p != "DATA_TYPE" and getattr(self, p, None) != policies[p]:
                    raise ValueError("Cannot change {} via PUT".format(p))

        if (
            getattr(self, "refresh_token", None) is None and not self.uses_service_auth
        ) or self.invalid:
            if not doc.get("invalid", True):
                raise ValueError("Cannot mark this credential as valid")
            doc["invalid"] = True
        elif "invalid" not in doc:
            doc["invalid"] = self.invalid

        super().put(**doc)

    def put_credentials(self, credentials):
        patch = self._credentials_to_patch(credentials)
        self._patch(**patch)
        self.refresh_configuration_hash()

    def _credentials_to_patch(self, credentials):
        if credentials:
            id_token = credentials.id_token
            refresh_token = credentials.refresh_token
            access_token = credentials.access_token

            patch = {
                "DATA_TYPE": self._DATA_TYPE,
                "display_name": id_token.get("name", id_token["email"]),
                "policies": {
                    "DATA_TYPE": self.POLICY_DATA_TYPE,
                    "sub": id_token["sub"],
                    "refresh_token": refresh_token,
                    "email": id_token["email"],
                    "access_token": access_token,
                    "token_expiry": credentials.token_expiry,
                    "scopes": credentials.scopes,
                },
                "invalid": False,
                "deleted": False,
            }
        else:
            patch = {
                "display_name": "{} (unprovisioned)".format(self.username),
                "policies": {
                    "DATA_TYPE": self.POLICY_DATA_TYPE,
                    "access_token": None,
                    "refresh_token": None,
                    "token_expiry": self.epoch,
                },
                "deleted": False,
            }
        patch["invalid"] = (
            patch["policies"]["refresh_token"] is None and not self.uses_service_auth
        )
        if not patch["invalid"]:
            patch["provisioned"] = True

        return patch

    @property
    def token_uri(self) -> str:
        raise AttributeError(
            "The token_uri property is not defined for this credential"
        )

    @property
    def supports_revoke(self) -> bool:
        return True

    @property
    def uses_service_auth(self) -> bool:
        return False
