from __future__ import annotations

import typing as t

import marshmallow as ma

from .data_type import Schema, data_type, data_type_schema


class StringOrObjectField(ma.fields.Field):
    """
    Polymorphic field. This can be used to de/serialize any object based on the
    DATA_TYPE field or a plain string.
    """

    def _jsonschema_type_mapping(self):
        return {}

    def _serialize(self, value, attr, obj, **kwargs):
        dt = ma.utils.get_value(value, "DATA_TYPE")
        if dt:
            schema = data_type_schema.schema_for(dt)
            return schema.dump(value)
        else:
            return value

    def _deserialize(self, value, attr, data, **kwargs):
        dt = ma.utils.get_value(value, "DATA_TYPE")
        if dt:
            schema = data_type_schema.schema_for(dt)
            return schema.load(value)
        else:
            return value


@data_type("result#1.0.0")
class Result_1_0_0(Schema):
    """
    This is the result envelope returned from all operations in this API. Each
    operation may add properties to this base document type with additional
    operation-specific data values.
    """

    http_response_code = ma.fields.Int(
        required=True,
        validate=[ma.validate.Range(100, 599)],
        description="Numeric HTTP response code",
    )
    code = ma.fields.String(
        required=True,
        description="String response code",
    )
    detail = StringOrObjectField(description="Machine readable response details")
    message = ma.fields.String(description="Message describing this result")
    data = ma.fields.Raw(type="array", items={"type": "object"})
    has_next_page = ma.fields.Boolean(
        default=False,
        description="""
        Boolean flag indicating whether or not additional pages of response
        data may be requested by passing the marker to the same operation.
        """,
    )

    marker = ma.fields.String(
        allow_none=True,
        description="""
        Opaque marker that may be passed to this API call to fetch the next
        page of results if the returned document has `has_next_page` set to
        true.
        """,
    )


@data_type("result#1.1.0")
class Result_1_1_0(Result_1_0_0):
    __doc__ = (
        t.cast(str, Result_1_0_0.__doc__)
        + """
    Version 1.1.0 adds optional authorization_parameters to help process
    authorization or authentication errors
    """
    )

    authorization_parameters = ma.fields.Dict(allow_none=True)


Result = data_type_schema.Result
