import marshmallow as ma

from ..api import (
    PolicySchema,
    collection_policy,
    data_type_schema,
    storage_gateway_policy,
    user_credential_policy,
)

BOX_CONNECTOR_ID = "7c100eae-40fe-11e9-95a3-9cb6d0d9fd63"


class BoxAppAuth(ma.Schema):
    """
    Key information used to perform JWT grants for using the Box API
    """

    publicKeyID = ma.fields.Str(required=True, description="ID of the public key")
    privateKey = ma.fields.Str(required=True, description="Private key")
    passphrase = ma.fields.Str(
        required=True, description="Passphrase to decrypt the private key"
    )


class BoxAppSettings(ma.Schema):
    """
    Values from the Box JWT client configuration that the storage gateway uses
    to identify and authenticate with the Box API.  This is only set when
    configuring the storage gateway for Box enterprise authentication.
    """

    clientID = ma.fields.Str(required=True, description="Application client ID")
    clientSecret = ma.fields.Str(required=True, description="Application client secret")
    appAuth = ma.fields.Nested(
        BoxAppAuth, required=True, description="Box application keys"
    )


class BoxJsonConfig(PolicySchema):
    boxAppSettings = ma.fields.Nested(
        BoxAppSettings,
        required=True,
        description="Box Application settings",
    )
    enterpriseID = ma.fields.Str(
        required=True,
        description="""
        Identifies which Box Enterprise this storage gateway is authorized
        access to. This is only set when configuring the storage gateway
        for Box enterprise authentication.
        """,
    )
    private = ["boxAppSettings", "enterpriseID"]


@storage_gateway_policy("box_storage_policies#1.0.0")
class BoxStorage_1_0_0(BoxJsonConfig):
    """
    Connector-specific storage gateway policies for the Box connector.
    """

    connector_id = BOX_CONNECTOR_ID


@storage_gateway_policy("box_storage_policies#1.1.0")
class BoxStorage_1_1_0(PolicySchema):
    """
    Connector-specific storage gateway policies for the Box connector.
    """

    boxAppSettings = ma.fields.Nested(BoxAppSettings, required=False)
    enterpriseID = ma.fields.Str(
        required=False,
        description="""
        Identifies which Box Enterprise this storage gateway is authorized
        access to. This is only set when configuring the storage gateway for
        Box enterprise authentication.
        """,
    )
    client_id = ma.fields.Str(
        required=False,
        description="""
        Client ID of the Box OAuth2 application registered on the Box developer
        console.  This is only set when configuring the storage gateway for
        OAuth2 user authentication.
        """,
    )
    secret = ma.fields.Str(
        required=False,
        description="""
        Secret associated with the client_id set in this policy.  This is only
        set when configuring the storage gateway for OAuth2 user
        authentication.
        """,
    )
    user_api_rate_limit = ma.fields.Int(
        required=False,
        description="""
        User API Rate Limit associated with this client ID in operations per
        second per user.
        """,
    )
    user_credential_required = ma.fields.Bool(
        dump_only=True,
        description="""
        Flag indicating whether users must register a credential.  If true,
        this storage gateway is configured for OAuth2 user authentication.  If
        false (and for older DATA_TYPE where this property is missing), this
        storage gateway is configured for enterprise authentication.
        """,
    )
    auth_callback = ma.fields.URL(
        schemes=["https"],
        dump_only=True,
        description="""
        URL of the auth callback that must be set on the Box developer
        console for the Box application of client_id.
        """,
    )

    private = [
        "client_id",
        "secret",
        "user_api_rate_limit",
        "boxAppSettings",
        "enterpriseID",
    ]
    connector_id = BOX_CONNECTOR_ID

    @ma.pre_load
    def ignore_dump_only(self, data, **kwargs):
        """
        Ignore dump_only schema fields when loading.
        """
        for k, v in self.fields.items():
            if v.dump_only:
                data.pop(k, None)
        return data

    @ma.validates_schema
    def old_or_new(self, data, **kwargs):
        # Either (boxAppSettings, enterpriseID) or (client_id, secret)
        partial = kwargs.get("partial", False)
        if partial:
            old_style = "boxAppSettings" in data and "enterpriseID" in data
            new_style = "client_id" in data and "secret" in data
            if not old_style and not new_style:
                return
        else:
            old_style = "boxAppSettings" in data or "enterpriseID" in data
            new_style = "client_id" in data or "secret" in data

        if not (old_style ^ new_style):
            raise ma.ValidationError(
                f"The {self._DATA_TYPE} document must contain either "
                f"boxAppSettings and enterpriseID or client_id and secret {kwargs}"
            )


@storage_gateway_policy("box_storage_policies#1.2.0")
class BoxStorage_1_2_0(BoxStorage_1_1_0):
    """
    Connector-specific storage gateway policies for the Box connector.
    """

    allow_any_account = ma.fields.Bool(
        required=False,
        load_default=False,
        description="""
        If true, allow users to access personal or external Box accounts.
        If false (the default), users must use the Box account which
        matches the username their Globus credential maps to.
        """,
    )

    connector_id = BOX_CONNECTOR_ID


@collection_policy("box_collection_policies#1.0.0")
class BoxCollectionPolicies_1_0_0(PolicySchema):
    """
    Connector-specific collection policies for the Box connector
    """

    connector_id = BOX_CONNECTOR_ID


@user_credential_policy("box_user_credential_policies#1.0.0")
class BoxUserCredential_1_0_0(PolicySchema):
    """
    Connector-specific user credential policies for the Box connector
    """

    connector_id = BOX_CONNECTOR_ID


@user_credential_policy("box_user_credential_policies#1.1.0")
class BoxUserCredential_1_1_0(PolicySchema):
    """
    Connector-specific user credential policies for the Box connector
    """

    sub = ma.fields.Str(allow_none=True, description="OAuth subject identifier claim")
    email = ma.fields.Email(allow_none=True, description="OAuth email identifier claim")
    access_token = ma.fields.Str(
        required=False, allow_none=True, description="OAuth access token"
    )
    refresh_token = ma.fields.Str(
        required=False, allow_none=True, description="OAuth refresh token"
    )
    token_expiry = ma.fields.AwareDateTime(
        description="OAuth access token expiration time"
    )
    scopes = ma.fields.List(
        ma.fields.Str,
        description="OAuth scopes associated with this access token",
        example=["box_readwrite"],
    )
    max_upload = ma.fields.Int()

    private = ["access_token", "refresh_token", "max_upload"]
    connector_id = BOX_CONNECTOR_ID


BoxCollectionPolicies = data_type_schema.BoxCollectionPolicies
BoxStoragePolicies = data_type_schema.BoxStoragePolicies
BoxUserCredentialPolicies = data_type_schema.BoxUserCredentialPolicies
