import typing as t

import marshmallow as ma

from ..api import (
    PolicySchema,
    collection_policy,
    data_type_schema,
    storage_gateway_policy,
    user_credential_policy,
)

S3_CONNECTOR_ID = "7643e831-5f6c-4b47-a07f-8ee90f401d23"


@storage_gateway_policy("s3_storage_policies#1.0.0")
class S3StoragePolicies_1_0_0(PolicySchema):
    __doc__ = """
    Connector-specific storage gateway policies for the S3 connector
    """

    s3_endpoint = ma.fields.Url(
        description="URL of the S3 API endpoint",
        example="https://s3.amazonaws.com",
    )
    s3_buckets = ma.fields.List(
        ma.fields.String,
        description="""
        List of buckets not owned by the collection owner that will be shown in
        the root of collections created at the base of this storage gateway.
        """,
    )
    s3_user_credential_required = ma.fields.Boolean(
        description="""
        Flag indicating if a Globus User must register a user credential in
        order to create a guest collection on this storage gateway.
        """
    )
    connector_id = S3_CONNECTOR_ID


@storage_gateway_policy("s3_storage_policies#1.1.0")
class S3StoragePolicies_1_1_0(S3StoragePolicies_1_0_0):
    __doc__ = (
        t.cast(str, S3StoragePolicies_1_0_0.__doc__)
        + """
    Version 1.1.0 adds support for the s3_requester_pays property
    """
    )

    s3_requester_pays = ma.fields.Boolean(
        description="""
        Flag indicating that S3 operations will be charged to the account of
        the registered credentials. Credentials used with a storage gateway
        that has the s3_requester_pays property set to true are invalid unless
        they also have this property set to true as an acknowledgement.
        """
    )
    connector_id = S3_CONNECTOR_ID


@collection_policy("s3_collection_policies#1.0.0")
class S3CollectionPolicies_1_0_0(PolicySchema):
    """
    Connector-specific collection policies for the S3 connector
    """

    connector_id = S3_CONNECTOR_ID


@user_credential_policy("s3_user_credential_policies#1.0.0")
class S3UserCredentialPolicies_1_0_0(PolicySchema):
    __doc__ = """
    Connector-specific user credential policies for the S3 connector
    """

    s3_key_id = ma.fields.String(
        allow_none=True,
        description="""
        Access Key ID to use with the S3 API to access your buckets and
        objects.
        """,
    )
    s3_secret_key = ma.fields.String(
        allow_none=True,
        description="""
        Secret key to use with the S3 API to access your buckets and objects.
        """,
    )

    private = ["s3_secret_key"]
    connector_id = S3_CONNECTOR_ID


@user_credential_policy("s3_user_credential_policies#1.1.0")
class S3UserCredentialPolicies_1_1_0(S3UserCredentialPolicies_1_0_0):
    __doc__ = (
        t.cast(str, S3UserCredentialPolicies_1_0_0.__doc__)
        + """
    Version 1.1.0 adds support for the s3_requester_pays property.
    """
    )

    s3_requester_pays = ma.fields.Boolean(
        description="""
        Flag indicating the user acknowledges S3 operations will be charged to
        the account of this credential.  If this flag is true in the storage
        gateway policy, this must also be true or the credential will be invalid.
        """
    )


S3CollectionPolicies = data_type_schema.S3CollectionPolicies
S3StoragePolicies = data_type_schema.S3StoragePolicies
S3UserCredentialPolicies = data_type_schema.S3UserCredentialPolicies
