from cachelib import SimpleCache
from flask import current_app

from globus_sdk.scopes import AuthScopes, TransferScopes

CLIENTS_API_BASE_URL = "/v2/api/clients"

cache = SimpleCache()


def _child_client_representation(collection):
    name = 'Collection "{}"'.format(collection.display_name)
    if len(name) > 100:
        name = name[:96] + '..."'
    vc = {
        "client": {
            "name": name,
            "public_client": False,
            "visibility": "public",
            "redirect_uris": ["https://{}/redirect".format(collection.domain_name)],
        }
    }

    return vc


def _scope_name_to_id(scope_name):
    cache_key = scope_name

    scope_id = cache.get(cache_key)
    if scope_id:
        return scope_id

    path = "/v2/api/scopes"
    query_params = {"scope_strings": scope_name}

    res = current_app.auth_client.get(path, query_params=query_params)

    scope_id = res.data["scopes"][0]["id"]
    cache.set(cache_key, scope_id, 7200)

    return scope_id


def _data_access_scope(collection, collection_client_id):
    display_name = "Manage your data on {}".format(collection.display_name)
    description = (
        "Allows synchronous access to your data, submission of asynchronous "
        "tasks for data management, and (if enabled) creation of guest "
        "collections on {} ({})".format(collection.display_name, collection.domain_name)
    )
    scope = {
        "name": display_name,
        "description": description,
        "scope_suffix": "data_access",
        "client": collection_client_id,
    }
    if current_app.tutorial_endpoint:
        allowed_domains = [
            d for d in collection.storage_gateway.allowed_domains if d != "*"
        ]
        if allowed_domains:
            scope["required_domains"] = allowed_domains
    else:
        scope["required_domains"] = collection.storage_gateway.allowed_domains
    return {"scope": scope}


def _https_scope(collection, collection_client_id):
    dependent_scopes = [
        {
            "scope": _scope_name_to_id(AuthScopes.openid),
            "optional": False,
            "requires_refresh_token": False,
            "allows_refresh_token": True,
        },
    ]

    if collection.is_guest_collection:
        dependent_scopes.append(
            {
                "scope": _scope_name_to_id(TransferScopes.all),
                "optional": False,
                "requires_refresh_token": False,
                "allows_refresh_token": True,
            }
        )
    display_name = "Access your data on {} via HTTPS".format(collection.display_name)
    description = "Access your data on {} ({}) via HTTPS".format(
        collection.display_name, collection.domain_name
    )

    return {
        "scope": {
            "name": display_name,
            "description": description,
            "scope_suffix": "https",
            "client": collection.id,
            "dependent_scopes": dependent_scopes,
        }
    }


def register_child_client(collection):
    res = current_app.auth_client.post(
        CLIENTS_API_BASE_URL, data=_child_client_representation(collection)
    )

    return res.data


def update_child_client(collection):
    res = current_app.auth_client.put(
        "{}/{}".format(CLIENTS_API_BASE_URL, collection.id),
        data=_child_client_representation(collection),
    )
    return res.data


def register_scopes(collection, collection_client_id):
    scopes = []
    data_access_scope = None
    if collection.is_mapped_collection and not collection.high_assurance:
        data_access_res = current_app.auth_client.post(
            "{}/{}/scopes".format(CLIENTS_API_BASE_URL, collection_client_id),
            data=_data_access_scope(collection, collection_client_id),
        )
        data_access_scope = data_access_res.data["scopes"][0]
        scopes.append(data_access_scope)

    https_res = current_app.auth_client.post(
        "{}/{}/scopes".format(CLIENTS_API_BASE_URL, collection_client_id),
        data=_https_scope(collection, collection_client_id),
    )
    https_scope = https_res.data["scopes"][0]
    scopes.append(https_scope)

    return {"scopes": scopes}


def register_fqdn(collection, collection_client_id):
    res = current_app.auth_client.post(
        "{}/{}/fqdns".format(CLIENTS_API_BASE_URL, collection_client_id),
        data={"fqdn": collection.domain_name},
    )

    return res.data


def register(collection):
    client = register_child_client(collection)
    collection.id = client["client"]["id"]
    # FIXME: register collections with auth once wildcards are supported
    # register_fqdn(collection, collection.id)
    scope = register_scopes(collection, collection.id)

    return client["client"], scope["scopes"]


def unregister(client_id):
    resource = "{}/{}".format(CLIENTS_API_BASE_URL, client_id)

    res = current_app.auth_client.delete(resource)

    return res.data
