import contextlib
import datetime
import os
from typing import Any, Callable, Dict, List, Optional, Union
from urllib.parse import urlparse

import marshmallow as ma
import sqlalchemy.orm as orm
from flask import current_app, g
from sqlalchemy.exc import IntegrityError

import globus_sdk
from globus_sdk.scopes import ScopeBuilder

import globus.manager.api.models as apimodels
from globus.manager.api.schemas.appsync.appsync_data import appsync_data_schema
from globus.manager.constants import ConfigurationObjectTag
from globus.manager.utils.interprocess_lock import InterprocessLock

from .query import (
    GraphQLArgument,
    GraphQLError,
    GraphQLQuery,
    GraphQLQueryType,
    GraphQLSelection,
    GraphQLVariable,
    GraphQLVariables,
)
from .subscription import Subscription


class AppsyncScopeBuilder(ScopeBuilder):
    @property
    def manager_configuration(self) -> str:
        return self.url_scope_string("manager_configuration")


class GraphQLResponse(globus_sdk.GlobusHTTPResponse):
    """
    Class used to parse results from the GraphQL service. Note that normally
    the GraphQL API returns errors within a HTTP 200 message response. This
    extracts some bits from the HTTP headers which may indicate some error
    that happens at a higher level than the GraphQL protocol.
    """

    def __init__(self, http_response, client=None):
        super().__init__(http_response, client=client)
        self.amazon_request_id = http_response._response.headers.get("x-amzn-RequestId")
        self.amazon_error_type = http_response._response.headers.get("x-amzn-ErrorType")
        self.amazon_message = self.data.get("message")


class GraphQLApiGatewayError(globus_sdk.GlobusAPIError):
    """
    GCS uses an API Gateway frontend to the GraphQL service. It can return
    some types of errors when parts of the system is down or some serious
    problems with the input (bad auth token, bad URL path).
    """

    def __init__(self, r):
        super().__init__(r)
        self.amazon_request_id = r.headers.get("x-amzn-RequestId")
        self.amazon_error_type = r.headers.get("x-amzn-ErrorType")


class UnexpectedApiGatewayMsgError(Exception):
    def __init__(self, endpoint: str, msg: str):
        self._endpoint = endpoint
        self._msg = msg

    def __str__(self):
        return f"{self._endpoint} endpoint returned an unexpected reply: {self._msg}"


class GraphQLClient(globus_sdk.BaseClient):
    model_order = [
        ConfigurationObjectTag.KEY_CHAIN,
        ConfigurationObjectTag.ENDPOINT,
        ConfigurationObjectTag.WILDCARD_DOMAIN,
        ConfigurationObjectTag.ACME_ACCOUNT,
        ConfigurationObjectTag.CERTIFICATE,
        ConfigurationObjectTag.STORAGE_GATEWAY,
        ConfigurationObjectTag.USER_CREDENTIAL,
        ConfigurationObjectTag.COLLECTION,
        ConfigurationObjectTag.ROLE,
    ]
    ENDPOINT_SELECTIONS = [
        "owner_id",
        "deleted",
        GraphQLSelection(
            "configuration_objects",
            selections=[
                GraphQLSelection(
                    "items",
                    selections=[
                        "id",
                        "tag",
                        "serial",
                        "created_at",
                        "updated_at",
                        "configuration",
                        "deleted",
                    ],
                ),
                "next_page_token",
            ],
        ),
    ]
    error_class = GraphQLApiGatewayError
    service_name = "appsync"

    def __init__(self, owner_id, *args, **kwargs) -> None:
        self.owner_id = owner_id
        self.subscription_lock: Optional[InterprocessLock]

        if "HOME" in current_app.config:
            self.subscription_lock = InterprocessLock(
                "subscription_pause", os.path.join(current_app.config["HOME"], "locks")
            )
        else:
            self.subscription_lock = None

        super().__init__(*args, **kwargs)

    def subscribe(self, **kwargs) -> Subscription:
        host: str = urlparse(self.base_url).netloc
        subscription_info = {
            "WSS_URL": f"wss://{host}/graphql/realtime",
            "HOST": host,
        }

        assert self.authorizer is not None
        access_token = self.authorizer.get_authorization_header()

        assert access_token is not None
        if access_token.startswith("Bearer "):
            access_token = access_token[len("Bearer ") :]

        return Subscription(
            subscription_info,
            access_token,
            self.owner_id,
            current_app._get_current_object(),
            **kwargs,
        )

    @contextlib.contextmanager
    def pause_notifications(self):
        """
        Pause subscription message processing
        """
        with self.subscription_lock.shared():
            yield

    def create_object(self, model):
        query_vars = []
        query_args = []

        owner_id_var = GraphQLVariable("owner_id", "ID!", self.owner_id)
        query_vars.append(owner_id_var)
        query_args.append(GraphQLArgument("owner_id", owner_id_var))

        id_var = GraphQLVariable("id", "ID!", model.id)
        query_vars.append(id_var)
        query_args.append(GraphQLArgument("id", id_var))

        tag_var = GraphQLVariable("tag", "ConfigurationObjectTag!", model.tag)
        query_vars.append(tag_var)
        query_args.append(GraphQLArgument("tag", tag_var))

        configuration_var = GraphQLVariable(
            "configuration", "String!", model.configuration
        )
        query_vars.append(configuration_var)
        query_args.append(GraphQLArgument("configuration", configuration_var))

        query_variables = GraphQLVariables(query_vars)

        query = GraphQLQuery(
            GraphQLQueryType.MUTATION,
            GraphQLSelection(
                "ObjectCreate",
                arguments=query_args,
                selections=self.ENDPOINT_SELECTIONS,
            ),
            query_variables,
        )

        res = self.post(
            "graphql",
            data=dict(query=str(query), variables=query_variables.variables()),
        )

        res = self._parse_result(res, "ObjectCreate")

        assert len(res["items"]) == 1
        for o in res["items"]:
            o["tag"] = ConfigurationObjectTag(o["tag"])
            model.update(o)

        return model

    def update_object(self, model, ignore_response=False, dependent_objs=None):
        query_vars = []
        query_args = []

        owner_id_var = GraphQLVariable("owner_id", "ID!", self.owner_id)
        query_vars.append(owner_id_var)
        query_args.append(GraphQLArgument("owner_id", owner_id_var))

        id_var = GraphQLVariable("id", "ID!", model.id)
        query_vars.append(id_var)
        query_args.append(GraphQLArgument("id", id_var))

        serial_var = GraphQLVariable("serial", "Int!", model.serial + 1)
        query_vars.append(serial_var)
        query_args.append(GraphQLArgument("serial", serial_var))

        if dependent_objs:
            dependents_var = GraphQLVariable(
                "dependents",
                "[DependentObject!]",
                [{"id": do.id, "serial": do.serial} for do in dependent_objs],
            )

            query_vars.append(dependents_var)
            query_args.append(GraphQLArgument("dependents", dependents_var))

        configuration_var = GraphQLVariable(
            "configuration", "String!", model.configuration
        )
        query_vars.append(configuration_var)
        query_args.append(GraphQLArgument("configuration", configuration_var))

        query_variables = GraphQLVariables(query_vars)

        query = GraphQLQuery(
            GraphQLQueryType.MUTATION,
            GraphQLSelection(
                "ObjectUpdate",
                arguments=query_args,
                selections=self.ENDPOINT_SELECTIONS,
            ),
            query_variables,
        )

        res = self.post(
            "graphql",
            data=dict(query=str(query), variables=query_variables.variables()),
        )

        if not ignore_response:
            res = self._parse_result(res, "ObjectUpdate")

            assert len(res["items"]) == 1
            for o in res["items"]:
                o["tag"] = ConfigurationObjectTag(o["tag"])
                model.update(o)

        return model

    def delete_object(self, model=None, id=None):
        query_vars = []
        query_args = []

        owner_id_var = GraphQLVariable("owner_id", "ID!", self.owner_id)
        query_vars.append(owner_id_var)
        query_args.append(GraphQLArgument("owner_id", owner_id_var))

        if model:
            id_val = model.id
        elif id is not None:
            id_val = id
        else:
            raise RuntimeError()

        id_var = GraphQLVariable("id", "ID!", id_val)
        query_vars.append(id_var)
        query_args.append(GraphQLArgument("id", id_var))

        query_variables = GraphQLVariables(query_vars)

        query = GraphQLQuery(
            GraphQLQueryType.MUTATION,
            GraphQLSelection(
                "ObjectDelete",
                arguments=query_args,
                selections=self.ENDPOINT_SELECTIONS,
            ),
            query_variables,
        )

        res = self.post(
            "graphql",
            data=dict(query=str(query), variables=query_variables.variables()),
        )

        res = self._parse_result(res, "ObjectDelete")
        return res

    def query_all(self, item_filter=None, since=None, tag=None, id=None):
        """
        Query AppSync for configuration information. This can be filtered
        by time, object tag, or id. This function process pagination
        internally.

        The since value can be either a numerical unix time offset or a
        datetime.datetime object. If it is a datetime.datetime and does not
        contain a timezone it is interpreted as a UTC time.
        """
        has_next_page = True
        next_page_token = None
        all_items = []
        iterations = 0

        while has_next_page:
            items, next_page_token = self.query(
                item_filter=item_filter,
                since=since,
                tag=tag,
                id=id,
                page_token=next_page_token,
            )
            all_items += items
            has_next_page = next_page_token
            iterations += 1

        current_app.logger.debug(
            "%s pages of data with %s items", iterations, len(all_items)
        )

        return all_items

    @staticmethod
    def _awsdatetime(since: Union[datetime.datetime, int, float]) -> str:
        if isinstance(since, (int, float)):
            since = datetime.datetime.fromtimestamp(since, datetime.timezone.utc)
        if not isinstance(since, datetime.datetime):
            raise TypeError("since must be numeric or datetime")
        offset = since.utcoffset()
        if offset is None:
            offset = datetime.timedelta(seconds=0)
        adjusted = (since + offset).replace(tzinfo=None)
        since_string = adjusted.isoformat() + "Z"
        return since_string

    def query(self, item_filter=None, since=None, tag=None, id=None, page_token=None):
        """
        Query AppSync for configuration information. This can be filtered
        by time, object tag, or id, and can be paginated.

        The since value can be either a numerical unix time offset or a
        datetime.datetime object. If it is a datetime.datetime and does not
        contain a timezone it is interpreted as a UTC time.

        If item_filter is set to a callable, it will be called with a list
        of responses to determine which ones to process into models to return
        """
        query_vars = []
        query_args = []

        if since is not None:
            since = self._awsdatetime(since)
            since_var = GraphQLVariable("since", "AWSDateTime", since)
            query_vars.append(since_var)
            query_args.append(GraphQLArgument("since", since_var))
        if tag is not None:
            if not isinstance(tag, ConfigurationObjectTag):
                tag = ConfigurationObjectTag(tag)
            tag_var = GraphQLVariable("tag", "ConfigurationObjectTag", tag)
            query_vars.append(tag_var)
            query_args.append(GraphQLArgument("tag", tag_var))

        if id is not None:
            id_var = GraphQLVariable("id", "ID", str(id))
            query_vars.append(id_var)
            query_args.append(GraphQLArgument("id", id_var))

        if page_token is not None:
            page_token_var = GraphQLVariable("page_token", "String", page_token)
            query_vars.append(page_token_var)
            query_args.append(GraphQLArgument("page_token", page_token_var))

        query_variables = GraphQLVariables(query_vars)

        query = GraphQLQuery(
            GraphQLQueryType.QUERY,
            GraphQLSelection(
                "Endpoint",
                selections=[
                    "owner_id",
                    "deleted",
                    GraphQLSelection(
                        "configuration_objects",
                        arguments=query_args,
                        selections=[
                            GraphQLSelection(
                                "items",
                                selections=[
                                    "owner_id",
                                    "id",
                                    "tag",
                                    "serial",
                                    "created_at",
                                    "updated_at",
                                    "configuration",
                                    "deleted",
                                ],
                            ),
                            "next_page_token",
                        ],
                    ),
                ],
            ),
            query_variables,
        )

        res = self.post(
            "graphql",
            data=dict(query=str(query), variables=query_variables.variables()),
        )

        res = self._parse_result(res, "Endpoint")

        if item_filter:
            res["items"] = item_filter(res["items"])

        return (
            res["items"],
            res.get("configuration_objects", {}).get("next_page_token"),
        )

    @staticmethod
    def _item_sort_key(item) -> int:
        sort_key = GraphQLClient.model_order.index(item["tag"]) * 10
        if item["tag"] == ConfigurationObjectTag.COLLECTION:
            configuration = current_app.keychain.decrypt(item["configuration"])
            # When an item is deleted from appsync, we will get an empty
            # configuration blob. In that case, we'll treat it as a guest
            # collection so it is deleted after mapped collections.
            collection_type = configuration.get("collection_type", "deleted")
            if collection_type == "guest":
                sort_key += 5
            elif collection_type == "deleted":
                sort_key += 7
        return sort_key

    @staticmethod
    def commit_items_to_session(
        session: orm.Session, items: List[Dict[str, Any]], ignore_old: bool = False
    ) -> List[Dict[str, Any]]:
        """
        After querying appsync, construct (or delete) the model objects based
        on the configuration_object items returned from appsync.

        :param session:
            The session to which this adds or deletes the models. The changes
            are committed in this function.

        :param items:
            A list of items from appsync.

        :param ignore_old:
            If true, ignore items which have an older serial number.

        :return:
            Returns a list of items which could not be safely added to the
            database because of a dangling foreign key constraint. They can be
            resubmitted this function along with later notification data.
        """
        deferred_items = []

        with apimodels.db.session.no_autoflush:
            previous_key = -1
            for item in sorted(items, key=GraphQLClient._item_sort_key):
                this_key = GraphQLClient._item_sort_key(item)
                if this_key != previous_key and session.dirty or session.new:
                    session.commit()
                model = GraphQLClient.process_item(item, ignore_old=ignore_old)
                if model:
                    if isinstance(model, apimodels.db.Model):
                        if item.get("deleted"):
                            session.delete(model)
                        else:
                            session.add(model)

                    if isinstance(model, apimodels.Collection):
                        # Handle collection domain collisions: The API enforces
                        # a constraint that the custom domains are unique, but
                        # there's a race if multiple GCS managers are updating
                        # different collections to the same domain. If this collision
                        # occurs, each node may end up with different state, but at
                        # least this will prevent permanent out of sync for other
                        # objects that are sent along with the update containing
                        # the collision.
                        try:
                            apimodels.db.session.commit()
                        except IntegrityError as e:
                            current_app.logger.error(
                                "Unable to commit %s %s: %s",
                                item.get("tag", "UNKNOWN OBJECT TAG"),
                                item.get("id", "UNKNOWN OBJECT ID"),
                                str(e),
                            )
                            apimodels.db.session.rollback()
                            deferred_items.append(item)
                elif not item.get("deleted"):
                    # Disregard deleted items that have no existing model in the database.
                    # Defer non-deleted items that have no existing model in the database.
                    deferred_items.append(item)
                previous_key = this_key

        if session.new or session.dirty or session.deleted:
            session.commit()

        return deferred_items

    @staticmethod
    def process_item(
        item: Dict[str, Any], *, catch_exceptions: bool = True, ignore_old: bool = False
    ) -> Optional[apimodels.ConfigurationObject]:
        """
        Returns the model object associated with item afer applying updates.

        Note: When item['deleted'] = True, appsync will nullify item['configuration']
        so helper functions will need to rely on other values in item to find and instantiate
        the old model.

        All helpers should handle the fact that appsync may send updates out of order in which
        case item may have dependencies that have not yet been processed.

        :param item:
            An item from appsync.

        :param catch_exceptions:
            If true, log exceptions and return None instead of propagating the
            exception.

        :param ignore_old:
            If true, ignore items which have an older serial number.

        :return:
            - None if item's dependent objects do not yet exist in the DB (new)
            - The updated model associated with item (exists)
            - None if item['deleted'] and no instance of the model exists in the DB (hard-deleted)
            - None if an error occurs and catch_exceptions=True
        """
        dispatch: Dict[
            ConfigurationObjectTag,
            Callable[
                [Dict[str, Any], bool, bool], Optional[apimodels.ConfigurationObject]
            ],
        ] = {
            ConfigurationObjectTag.ACME_ACCOUNT: GraphQLClient._process_acme_account_item,
            ConfigurationObjectTag.CERTIFICATE: GraphQLClient._process_certificate_item,
            ConfigurationObjectTag.COLLECTION: GraphQLClient._process_collection_item,
            ConfigurationObjectTag.ENDPOINT: GraphQLClient._process_endpoint_item,
            ConfigurationObjectTag.KEY_CHAIN: GraphQLClient._process_keychain_item,
            ConfigurationObjectTag.ROLE: GraphQLClient._process_role_item,
            ConfigurationObjectTag.STORAGE_GATEWAY: GraphQLClient._process_storage_gateway_item,
            ConfigurationObjectTag.USER_CREDENTIAL: GraphQLClient._process_user_credential_item,
            ConfigurationObjectTag.WILDCARD_DOMAIN: GraphQLClient._process_wildcard_domain_item,
        }
        return dispatch[item["tag"]](item, catch_exceptions, ignore_old)

    @staticmethod
    def _process_item(
        item, model_class, catch_exceptions: bool, ignore_old: bool, **init_kwargs
    ):
        old_model = model_class.query.filter(model_class.id == item["id"]).one_or_none()

        if old_model:
            if ignore_old:
                if (
                    old_model.serial > item["serial"]
                    and old_model.created_at == item["created_at"]
                ):
                    return old_model

            try:
                if item.get("deleted", False):
                    old_model.deleted = True
                else:
                    old_model.update(item)
                return old_model
            except Exception as e:
                current_app.logger.error(
                    "Error processing update to object %s: %s", item["id"], str(e)
                )
                if not catch_exceptions:
                    raise
        else:
            try:
                if item.get("deleted", False):
                    current_app.logger.debug(
                        f"Ignoring deleted {item['tag']} item {item['id']}."
                    )
                else:
                    return model_class(appsync_data=item, **init_kwargs)
            except Exception as e:
                current_app.logger.info(
                    "Error processing previously unknown object %s: %s",
                    item["id"],
                    str(e),
                )
                if not catch_exceptions:
                    raise

    @staticmethod
    def _process_acme_account_item(item, catch_exceptions: bool, ignore_old: bool):
        return GraphQLClient._process_item(
            item,
            apimodels.AcmeAccount,
            catch_exceptions,
            ignore_old,
        )

    @staticmethod
    def _process_certificate_item(item, catch_exceptions: bool, ignore_old: bool):
        return GraphQLClient._process_item(
            item,
            apimodels.LetsEncryptCertificate,
            catch_exceptions,
            ignore_old,
        )

    @staticmethod
    def _get_configuration(
        item: Dict[str, Any], catch_exceptions: bool
    ) -> Optional[Dict[str, str]]:
        """
        Return the configuration_hash associated with item, either directly from
        hash or from the DB if the item has been deleted.

        This function does not perform any reconciling between old and new configuration
        values. Its purpose is to provide the caller with a useful configuration that
        can be used to determine static information like the IDs of dependent objects
        or the type of connector used.

        The caller can imply that a non-None return value means either the item is not
        deleted (dependent objects may or may not exist yet) or the item is deleted but
        the item object exists in the DB (its dependent objects must exist).
        """
        # Non-deleted items should have non-empty configurations
        if not item["deleted"]:
            try:
                return current_app.keychain.decrypt(item["configuration"])
            except Exception:
                current_app.logger.error(
                    "Error decrypting configuration %s", item["id"]
                )
                if not catch_exceptions:
                    raise
                return None

        # item['deleted'] will have an empty item['configuration']. Try to load the old
        # configuration object model. We're only after the configuration_hash.
        configuration_object: Optional[
            apimodels.ConfigurationObject
        ] = apimodels.ConfigurationObject.query.filter(
            apimodels.ConfigurationObject.id == item["id"]
        ).one_or_none()

        # This is expected if the object has already been hard-deleted
        if not configuration_object:
            current_app.logger.debug(
                f"No existing configuration for deleted {item['tag']} item {item['id']}. Ignoring."
            )
            return None
        # Should we copy this out?
        return configuration_object.configuration_hash

    @staticmethod
    def _process_collection_item(item, catch_exceptions: bool, ignore_old: bool):
        init_kwargs = {}
        configuration = GraphQLClient._get_configuration(item, catch_exceptions)
        if not configuration:
            return

        init_kwargs["endpoint"] = apimodels.Endpoint.query.one()

        storage_gateway_id = configuration.get("storage_gateway_id")
        if not storage_gateway_id:
            current_app.logger.error(f"No storage_gateway_id for item {item['id']}")
            return

        storage_gateway = apimodels.StorageGateway.query.filter(
            apimodels.StorageGateway.id == storage_gateway_id
        ).one_or_none()
        if not storage_gateway:
            # storage gateway item has not been processed yet
            current_app.logger.debug(
                f"Deferring collection item {item['id']}, "
                f"storage gateway {storage_gateway_id} does not resolve"
            )
            return None
        init_kwargs["storage_gateway"] = storage_gateway

        mapped_collection_id = configuration.get("mapped_collection_id")
        if mapped_collection_id:
            mapped_collection = apimodels.Collection.query.filter(
                apimodels.Collection.id == mapped_collection_id
            ).one_or_none()
            if not mapped_collection:
                current_app.logger.debug(
                    f"Deferring collection item {item['id']}, "
                    f"mapped collection {mapped_collection_id} does not resolve"
                )
                return None
            init_kwargs["mapped_collection"] = mapped_collection

        user_credential_id = configuration.get("user_credential_id")
        if user_credential_id:
            user_credential = apimodels.UserCredential.query.filter(
                apimodels.UserCredential.id == user_credential_id
            ).one_or_none()
            if not user_credential:
                current_app.logger.debug(
                    f"Deferring collection item {item['id']}, "
                    f"user credential {user_credential_id} does not resolve"
                )
                return None
            init_kwargs["user_credential"] = user_credential

        model_cls = apimodels.Collection.polymorphic_class(storage_gateway.connector_id)

        return GraphQLClient._process_item(
            item, model_cls, catch_exceptions, ignore_old, **init_kwargs
        )

    @staticmethod
    def _process_endpoint_item(item, catch_exceptions: bool, ignore_old: bool):
        return GraphQLClient._process_item(
            item, apimodels.Endpoint, catch_exceptions, ignore_old
        )

    @staticmethod
    def _process_keychain_item(item, catch_exceptions, ignore_old):
        try:
            current_app.keychain.update(item)
            return current_app.keychain
        except Exception as e:
            current_app.logger.error(
                "Error processing keychain %s: %s", item["id"], str(e)
            )
            if not catch_exceptions:
                raise

    @staticmethod
    def _process_role_item(item, catch_exceptions: bool, ignore_old: bool):
        configuration = GraphQLClient._get_configuration(item, catch_exceptions)
        if not configuration:
            return

        collection_id = configuration.get("resource")
        if collection_id:
            if not item.get("deleted"):
                collection = apimodels.Collection.query.filter(
                    apimodels.Collection.id == collection_id
                ).one_or_none()
                if not collection:
                    current_app.logger.debug(
                        "Deferring %s %s, collection item %s does not resolve",
                        item["tag"],
                        item["id"],
                        collection_id,
                    )
                    return None

        return GraphQLClient._process_item(
            item, apimodels.Role, catch_exceptions, ignore_old
        )

    @staticmethod
    def _process_storage_gateway_item(item, catch_exceptions: bool, ignore_old: bool):
        configuration = GraphQLClient._get_configuration(item, catch_exceptions)
        if not configuration:
            return

        model_cls = apimodels.StorageGateway.polymorphic_class(
            configuration["connector_id"]
        )

        return GraphQLClient._process_item(
            item, model_cls, catch_exceptions, ignore_old
        )

    @staticmethod
    def _process_user_credential_item(item, catch_exceptions: bool, ignore_old: bool):
        init_kwargs = {}
        configuration = GraphQLClient._get_configuration(item, catch_exceptions)
        if not configuration:
            return

        storage_gateway_id = configuration.get("storage_gateway_id")
        if not storage_gateway_id:
            current_app.logger.error(f"No storage_gateway_id for item {item['id']}")
            return

        storage_gateway = apimodels.StorageGateway.query.filter(
            apimodels.StorageGateway.id == storage_gateway_id
        ).one_or_none()
        if not storage_gateway:
            current_app.logger.debug(
                f"Deferring credential item {item['id']}, "
                f"storage gateway {storage_gateway_id} does not resolve"
            )
            return None
        init_kwargs["storage_gateway"] = storage_gateway

        model_cls = apimodels.UserCredential.polymorphic_class(
            storage_gateway.connector_id
        )

        return GraphQLClient._process_item(
            item, model_cls, catch_exceptions, ignore_old, **init_kwargs
        )

    @staticmethod
    def _process_wildcard_domain_item(item, catch_exceptions: bool, ignore_old: bool):
        return GraphQLClient._process_item(
            item,
            apimodels.GlobusDnsDomain,
            catch_exceptions,
            ignore_old,
        )

    def _parse_result(self, res, wrapper_property, unknown=ma.RAISE):
        res = GraphQLResponse(res)
        data = res.get("data", {}).get(wrapper_property, {})
        if data:
            try:
                data = appsync_data_schema.load(data, unknown=unknown)
            except Exception as e:
                current_app.logger.error(
                    "Error parsing result from appsync: %s", str(e)
                )

        if res.get("errors"):
            exc = GraphQLError(res["errors"], data)
            request_id = getattr(g, "request_id", None) if g else None

            current_app.logger.info(
                "graphql_exception: gcs_request_id=%s "
                'graphql_request_id=%s error=%s message="%s"',
                request_id,
                res.amazon_request_id,
                exc.errors[0]["errorType"],
                exc.message,
            )
            raise exc

        return data
