import datetime
import os

import globus.manager.api.models as models
from globus.manager.utils.interprocess_lock import InterprocessLock


class CertificateManager:
    RENEW_WINDOW = datetime.timedelta(days=30)
    MINIMUM_LEASE = datetime.timedelta(hours=1)
    CHECK_PERIOD = 3 * 24 * 60 * 60

    def __init__(self, scheduler, app):
        self.scheduler = scheduler
        self.app = app
        self.lock = InterprocessLock(
            "certificate", lockdir=os.path.join(app.config["HOME"], "locks")
        )

        check = self.CHECK_PERIOD
        with self.app.app_context(), self.lock.exclusive():
            now = datetime.datetime.now(tz=datetime.timezone.utc)
            for c in models.LetsEncryptCertificate.query:
                if (
                    now + datetime.timedelta(seconds=self.CHECK_PERIOD)
                ) > c.expiration_time:
                    check = 0
        self.scheduler.enter(check, 10, self.check_certificates)

    def check_certificates(self):
        self.scheduler.enter(self.CHECK_PERIOD, 10, self.check_certificates)

        with self.app.app_context(), self.lock.exclusive():
            acme_account = models.AcmeAccount.query.one()
            now = datetime.datetime.now(datetime.timezone.utc)

            have_lease = False
            for c in models.LetsEncryptCertificate.query.all():
                if now > (c.expiration_time - self.RENEW_WINDOW):
                    lease_remaining = acme_account.lease_remaining()
                    if lease_remaining <= self.MINIMUM_LEASE:
                        have_lease = self.refresh_lease(acme_account)
                    else:
                        have_lease = True

                    if have_lease:
                        self.renew_cert(c)

    def renew_cert(self, cert):
        try:
            self.app.logger.info("Renewing %s", cert)
            cert.renew()
            models.db.session.add(cert)
            self.app.appsync_client.update_object(cert)
            models.db.session.commit()
        except Exception as e:
            self.app.logger.error("Error renewing %s: %s", cert, str(e))
            models.db.session.rollback()
        cert.write_config()

    def refresh_lease(self, acme_account):
        try:
            self.app.logger.debug("Acquiring lease")
            acme_account.acquire_lease()
            models.db.session.add(acme_account)
            self.app.appsync_client.update_object(acme_account)
            models.db.session.commit()
            return True
        except Exception as e:
            self.app.logger.error("Error acquiring lease: %s", str(e))

            try:
                lease_owner = acme_account.lease_node_id
                lease_expiration = acme_account.lease_expiration
                self.app.logger.debug(
                    "Node %s has lease until %s", lease_owner, lease_expiration
                )
            except Exception as e:
                self.app.logger.error("Unable to acquire lease: %s", str(e))
            models.db.session.rollback()

        return False
