import json
import os

import click
from sqlalchemy.exc import OperationalError

from globus.manager.api.app import make_app
from globus.manager.api.models import Configuration, Endpoint, db
from globus.manager.api.schemas.appsync.oidc_registration import (
    OIDCRegistration as oidc_registration_schema,
)
from globus.manager.api.schemas.appsync.oidc_server import oidc_server_yaml_schema
from globus.manager.scripts.utils.setup_utils import load_keychain
from globus.manager.utils.keychain import NodeKey


def oidc_show(ctx):
    if os.getuid() != 0:
        click.echo("This command must be run as the root user", err=True)
        ctx.exit(1)

    app = make_app(app_config="globus.manager.api.config.oidc_cli")
    db.init_app(app)

    with app.app_context():
        try:
            for c in Configuration.query.all():
                if c.name == "node_key":
                    app.config[c.name.upper()] = NodeKey(**json.loads(c.value))
                else:
                    app.config[c.name.upper()] = c.value
        except OperationalError:
            raise click.ClickException(
                "This command must be run on a configured GCSv5 node"
            )
        try:
            # Load the keychain
            load_keychain()

            # Get the endpoint
            endpoint_model = Endpoint.query.one()

            if endpoint_model.has_oidc_server:
                oidc = oidc_server_yaml_schema.dump(endpoint_model.oidc_server)
                click.echo(
                    "Current OIDC server configuration: \r\n{}".format(
                        json.dumps(oidc, indent=2, sort_keys=True)
                    )
                )
            elif endpoint_model.has_oidc_registration:
                oidc = oidc_registration_schema.dump(endpoint_model.oidc_registration)
                click.echo(
                    "Current OIDC server registration: \r\n{}".format(
                        json.dumps(oidc, indent=2, sort_keys=True)
                    )
                )
            else:
                click.echo(
                    "An OIDC server has not been configured for this endpoint. "
                    "Run 'globus-connect-server oidc create' or "
                    "'globus-connect-server oidc register' to configure it."
                )
                ctx.exit(1)
        except Exception as e:
            click.echo("Error while retrieving OIDC endpoint: {}".format(str(e)))
            ctx.exit(1)
