"""
Utilities to parse and validate X.509 certificates and keys
"""
import datetime
import re
import ssl
from typing import Optional, Union

from OpenSSL import crypto


def _cert_to_ssl_format(cert: crypto.X509):
    """
    Helper to convert an X.509 certificate object into something that resembles
    the return value from getpeercert(). We really only care about
    subjectAltName, but we could get some obsolete certificate with only a
    subject name, so we parse the cert as best we can
    """
    aliases = {
        "C": "countryName",
        "ST": "stateOrProvinceName",
        "L": "localityName",
        "O": "organizationName",
        "OU": "organizationalUnitName",
        "CN": "commonName",
        "emailAddress": "emailAddress",
    }

    subject = tuple(
        [
            tuple(
                [
                    (aliases.get(k.decode(), k.decode()), v.decode())
                    for k, v in cert.get_subject().get_components()
                ]
            )
        ]
    )
    issuer = tuple(
        [
            [
                (aliases.get(k.decode(), k.decode()), v.decode())
                for k, v in cert.get_issuer().get_components()
            ]
        ]
    )
    notBefore_str = cert.get_notBefore()
    notBefore: Optional[str] = None
    if notBefore_str:
        notBefore = datetime.datetime.strptime(
            notBefore_str.decode("ascii"), "%Y%m%d%H%M%SZ"
        ).strftime("%b %d %H:%M:%S %Y GMT")
    notAfter_str = cert.get_notAfter()
    notAfter: Optional[str] = None
    if notAfter_str:
        notAfter = datetime.datetime.strptime(
            notAfter_str.decode("ascii"), "%Y%m%d%H%M%SZ"
        ).strftime("%b %d %H:%M:%S %Y GMT")
    else:
        notAfter = None

    certdata = {
        "subject": subject,
        "issuer": issuer,
        "serialNumber": cert.get_serial_number(),
        "version": cert.get_version(),
        "notBefore": notBefore,
        "notAfter": notAfter,
    }

    ext_count = cert.get_extension_count()
    extensions_data = {}
    for i in range(ext_count):
        ext = cert.get_extension(i)
        name = ext.get_short_name().decode()
        if name == "subjectAltName":
            val = [tuple(item.strip().split(":", 1)) for item in str(ext).split(",")]  # type: ignore
        else:
            val = str(ext)  # type: ignore
        extensions_data[name] = val

    certdata.update(extensions_data)

    return certdata


def parse_certificate(data: Union[str, bytes], hostname: Optional[str] = None) -> str:
    """
    Parse a string or byte stream as an X.509 certificate in PEM form, then
    return just the certificate data, eliminating any other text that might be
    in it.

    Optionally, if the hostname parameter is set, the certificate is compared
    against the hostname to verify that it will match.

    :param data:
        The raw PEM data to parse.
    :param hostname:
        Optional hostname to compare against this certificate.
    """
    if isinstance(data, str):
        data = data.encode("ascii")
    cert = crypto.load_certificate(crypto.FILETYPE_PEM, data)

    if hostname:
        certdata = _cert_to_ssl_format(cert)
        ssl.match_hostname(certdata, hostname)

    return crypto.dump_certificate(crypto.FILETYPE_PEM, cert).decode("ascii")


def parse_certificate_chain(data: Optional[Union[str, bytes]]) -> Optional[str]:
    """
    Parse an optional certificate chain from a str or bytes, discarding the
    non-PEM data. Returns either a string containing the PEM data, or None
    if there is no data.

    :param data:
        str or bytes containing the data to process.
    """
    chain_items = []

    start = r"-----BEGIN CERTIFICATE-----[\r]?\n"
    end = r"-----END CERTIFICATE-----[\r]?[\n]?"

    if isinstance(data, str):
        _data = data
    elif isinstance(data, bytes):
        _data = data.decode("ascii")
    elif data is None:
        return None

    while _data:
        startp = re.search(start, _data, re.MULTILINE)
        endp = re.search(end, _data, re.MULTILINE)
        if startp and endp and startp.start() < endp.end():
            chain_items.append(parse_certificate(_data[startp.start() : endp.end()]))
            _data = _data[endp.end() :]
        else:
            break

    if chain_items:
        return "\n".join(chain_items)
    return None


def parse_private_key(data: Union[str, bytes], callback=None) -> str:
    """
    Parse a private key in PEM format from a str or bytes, returning a str
    containing the PEM-encoded key.

    :param data:
        str or bytes contianing the PEM-encoded data
    :callback:
        str or bytes or callable to decrypt the private key.
    """
    private_key = crypto.load_privatekey(crypto.FILETYPE_PEM, data, passphrase=callback)
    return crypto.dump_privatekey(crypto.FILETYPE_PEM, private_key).decode("ascii")
