import hashlib
import math

STARTING_POINT = 0x1A  # non zero starting point with two hex digits


def _generate_xor(byte, hex_len):
    """
    byte: a positive int <= 255 (8 bits)
    hex_len: an it >= 2 for the desired hex_len of the xor
    return: an int intended to be used as an xor for masking/unmasking
    a number with hex_len digits in hexadecimal.

    the xor is generated by repeatedly xoring with the byte shifted left
    4 bits at a time, until the xor is hex_len nibbles long.
    """
    assert isinstance(byte, int)
    assert isinstance(hex_len, int)
    assert 0 <= byte <= 255
    assert 2 <= hex_len

    xor = 0
    for i in range(hex_len - 1):
        xor ^= byte << (4 * i)
    return xor


def obfuscate(n):
    """
    n: A number

    return: a string at least 4 chars long
        format: XXHH...
        X = hex of byte used to generate the xor
        H = hex of n shifted and xored

    examples starting at 0:
    5f45
    677c
    c2da
    b7ae
    627c

    examples starting at 1000:
    7a428
    3d01e
    9aaca
    6a53b
    dfed9
    """
    assert isinstance(n, int)
    assert n >= 0
    hex_len = max(math.ceil(n.bit_length() / 4), 2)

    # obfuscate our starting point (0)
    n ^= STARTING_POINT

    # generate an xor byte
    xor_byte = hashlib.sha256(str(n).encode("utf-8")).digest()[0]

    # obfuscate our sequential counting with an xor mask
    xor = _generate_xor(xor_byte, hex_len)
    n ^= xor

    # return formatted string
    format_string = "%%02x%%0%dx" % hex_len
    ret = format_string % (xor_byte, n)
    assert len(ret) == hex_len + 2
    return ret


def deobfuscate(s):
    """
    Given a string s created using the above obfuscate function
    Returns the original obfuscated int.
    """
    xor_byte = int(s[:2], 16)
    n = int(s[2:], 16)
    hex_len = len(s[2:])

    # unmask n
    xor = _generate_xor(xor_byte, hex_len)
    n ^= xor

    # de-obfuscate our starting point
    n ^= STARTING_POINT

    return n
