from __future__ import annotations

import os
import pwd
import typing as t
from pathlib import Path

from flask import render_template

from ..service_manager import ServiceManager
from ..utils import (
    log,
    write_and_link_file,
)

if t.TYPE_CHECKING:
    import globus.manager.api.models as models


PRIVATE_DATA_WARNING = """
******************************************************************************
WARNING: Your GridFTP server is behind a NAT, but the Server name resolves
to a private IP address. This probably won't work correctly with Globus.
******************************************************************************
"""


class GridFTPService:
    """
    Class to configure a GridFTP server
    """

    service: t.ClassVar[str] = "globus-gridftp-server"
    config_owner: t.ClassVar[str] = "gcsweb"

    etc_gridftp_d: t.ClassVar[Path] = Path("/etc/gridftp.d")
    var_gridftp_d: t.ClassVar[Path] = Path("/var/lib/globus-connect-server/gridftp.d")
    conf_file_name: t.ClassVar[Path] = var_gridftp_d / "globus-connect-server"
    conf_link_name: t.ClassVar[Path] = etc_gridftp_d / "globus-connect-server"

    def __init__(
        self,
        *,
        endpoint: models.Endpoint | None = None,
        node: models.Node | None = None,
        environment: str = "default",
        service_manager: t.Any | None = None,
        root: str = "/",
    ) -> None:
        if service_manager is None:
            service_manager = ServiceManager
        self.service_manager = service_manager
        self.endpoint = endpoint
        self.environment = environment
        self.node = node
        self.gridftp_key = Path(f"~{self.config_owner}/gridftp-key").expanduser()
        self.vhost_dir = Path(f"~{self.config_owner}/etc/domains").expanduser()

    def update(self, force: bool = False) -> None:
        if (
            self.node
            and hasattr(self.node, "data_interface")
            and self.node.data_interface.is_private
        ):
            log.warn(PRIVATE_DATA_WARNING)

        data = render_template(
            "gridftp_config.jinja2",
            endpoint=self.endpoint,
            node=self.node,
            environment=self.environment,
            os=os,
            gridftp_key=self.gridftp_key,
            vhost_dir=self.vhost_dir,
        )

        write_and_link_file(str(self.conf_file_name), str(self.conf_link_name), data)
        pwent = pwd.getpwnam(self.config_owner)
        os.chown(self.conf_file_name, pwent.pw_uid, pwent.pw_gid)

    def setup(self, force=False):
        self.update(force)

        self.service_manager.enable(self.service)
        if self.service_manager.status(self.service):
            self.service_manager.restart(self.service)
        else:
            self.service_manager.start(self.service)

    def cleanup(self, **kwargs):
        self.service_manager.stop(self.service)
        self.service_manager.disable(self.service)

        for name in os.listdir(self.etc_gridftp_d):
            if (
                name.startswith("globus-connect-server")
                and name != "globus-connect-server-package"
            ):
                os.remove(self.etc_gridftp_d / name)
        for name in os.listdir(self.var_gridftp_d):
            os.remove(self.var_gridftp_d / name)
